from pyasn1.type.univ import ObjectIdentifier


# top-level OID arcs

nistAlgorithms = ObjectIdentifier("2.16.840.1.101.3.4")

sigAlgs = nistAlgorithms + (3,)

kems = nistAlgorithms + (4,)

# ML-DSA

id_ml_dsa_44 = sigAlgs + (17,)

id_ml_dsa_65 = sigAlgs + (18,)

id_ml_dsa_87 = sigAlgs + (19,)

MLDSA_OIDS = {
    id_ml_dsa_44,
    id_ml_dsa_65,
    id_ml_dsa_87,
}

# HashML-DSA

id_hash_ml_dsa_44_with_sha512 = sigAlgs + (32,)

id_hash_ml_dsa_65_with_sha512 = sigAlgs + (33,)

id_hash_ml_dsa_87_with_sha512 = sigAlgs + (34,)

HASH_MLDSA_OIDS = {
    id_hash_ml_dsa_44_with_sha512,
    id_hash_ml_dsa_65_with_sha512,
    id_hash_ml_dsa_87_with_sha512,
}

# SLH-DSA

id_slh_dsa_sha2_128s = sigAlgs + (20,)

id_slh_dsa_sha2_128f = sigAlgs + (21,)

id_slh_dsa_sha2_192s = sigAlgs + (22,)

id_slh_dsa_sha2_192f = sigAlgs + (23,)

id_slh_dsa_sha2_256s = sigAlgs + (24,)

id_slh_dsa_sha2_256f = sigAlgs + (25,)

id_slh_dsa_shake_128s = sigAlgs + (26,)

id_slh_dsa_shake_128f = sigAlgs + (27,)

id_slh_dsa_shake_192s = sigAlgs + (28,)

id_slh_dsa_shake_192f = sigAlgs + (29,)

id_slh_dsa_shake_256s = sigAlgs + (30,)

id_slh_dsa_shake_256f = sigAlgs + (31,)

SLHDSA_OIDS = {
    id_slh_dsa_sha2_128s,
    id_slh_dsa_sha2_128f,
    id_slh_dsa_sha2_192s,
    id_slh_dsa_sha2_192f,
    id_slh_dsa_sha2_256s,
    id_slh_dsa_sha2_256f,
    id_slh_dsa_shake_128s,
    id_slh_dsa_shake_128f,
    id_slh_dsa_shake_192s,
    id_slh_dsa_shake_192f,
    id_slh_dsa_shake_256s,
    id_slh_dsa_shake_256f,
}

# HashSLH-DSA

id_hash_slh_dsa_sha2_128s_with_sha256 = sigAlgs + (35,)

id_hash_slh_dsa_sha2_128f_with_sha256 = sigAlgs + (36,)

id_hash_slh_dsa_sha2_192s_with_sha512 = sigAlgs + (37,)

id_hash_slh_dsa_sha2_192f_with_sha512 = sigAlgs + (38,)

id_hash_slh_dsa_sha2_256s_with_sha512 = sigAlgs + (39,)

id_hash_slh_dsa_sha2_256f_with_sha512 = sigAlgs + (40,)

id_hash_slh_dsa_shake_128s_with_shake128 = sigAlgs + (41,)

id_hash_slh_dsa_shake_128f_with_shake128 = sigAlgs + (42,)

id_hash_slh_dsa_shake_192s_with_shake256 = sigAlgs + (43,)

id_hash_slh_dsa_shake_192f_with_shake256 = sigAlgs + (44,)

id_hash_slh_dsa_shake_256s_with_shake256 = sigAlgs + (45,)

id_hash_slh_dsa_shake_256f_with_shake256 = sigAlgs + (46,)

HASH_SLHDSA_OIDS = {
    id_hash_slh_dsa_sha2_128s_with_sha256,
    id_hash_slh_dsa_sha2_128f_with_sha256,
    id_hash_slh_dsa_sha2_192s_with_sha512,
    id_hash_slh_dsa_sha2_192f_with_sha512,
    id_hash_slh_dsa_sha2_256s_with_sha512,
    id_hash_slh_dsa_sha2_256f_with_sha512,
    id_hash_slh_dsa_shake_128s_with_shake128,
    id_hash_slh_dsa_shake_128f_with_shake128,
    id_hash_slh_dsa_shake_192s_with_shake256,
    id_hash_slh_dsa_shake_192f_with_shake256,
    id_hash_slh_dsa_shake_256s_with_shake256,
    id_hash_slh_dsa_shake_256f_with_shake256,
}

# ML-KEM

id_alg_ml_kem_512 = kems + (1,)

id_alg_ml_kem_768 = kems + (2,)

id_alg_ml_kem_1024 = kems + (3,)

MLKEM_OIDS = {
    id_alg_ml_kem_512,
    id_alg_ml_kem_768,
    id_alg_ml_kem_1024,
}



from pkilint import finding_filter
from pkilint.cabf.serverauth import serverauth_subscriber
from pkilint.etsi.asn1 import ts_119_495 as ts_119_495_asn1


class Psd2CabfServerauthValidityPeriodFilter(finding_filter.FindingDescriptionFilter):
    _TARGET_VALIDATIONS = {
        serverauth_subscriber.SubscriberValidityPeriodValidator.VALIDATION_VALIDITY_PERIOD_EXCEEDS_397_DAYS,
        serverauth_subscriber.SubscriberValidityPeriodValidator.VALIDATION_VALIDITY_PERIOD_EXCEEDS_398_DAYS,
    }

    def filter(self, result, finding_description):
        if finding_description.finding in self._TARGET_VALIDATIONS:
            """
            OVR-6.1-3: TSPs issuing certificates for EU PSD2 may use the following policy identifier to augment the
            policy requirements associated with policy identifier QEVCP-w or QNCP-w as specified in
            ETSI EN 319 411-2 [5] giving precedence to the requirements defined in the present document.
            """
            return ts_119_495_asn1.qcp_web_psd2 not in result.node.document.policy_oids
        else:
            return True



import base64
import enum
import re

from pkilint.pkix.certificate import RFC5280Certificate
from pkilint.pkix.crl import RFC5280CertificateList
from pkilint.pkix.ocsp import RFC6960OCSPResponse


class DocumentFormat(enum.Enum):
    DETECT = enum.auto()
    BASE64 = enum.auto()
    DER = enum.auto()
    PEM = enum.auto()


class DocumentLoader:
    def __init__(self, document_cls, document_pem_label: str):
        self._document_cls = document_cls
        self._document_pem_label = document_pem_label.upper()

        self._pem_re = self._create_pem_re()

    def _create_pem_re(self) -> re.Pattern:
        ascii_armor_start = f"-----BEGIN {self._document_pem_label}-----"
        ascii_armor_end = f"-----END {self._document_pem_label}-----"

        return re.compile(
            f"^\\s*{ascii_armor_start}(?P<pem>.+){ascii_armor_end}\\s*$", re.DOTALL
        )

    def load_der_document(
        self,
        substrate: bytes,
        document_name: str = None,
        substrate_source: str = None,
        parent=None,
    ):
        if not substrate.startswith(b"\x30"):
            raise ValueError("Substrate is not DER-encoded")

        doc = self._document_cls(substrate_source, substrate, document_name, parent)
        doc.decode()

        return doc

    def load_der_file(
        self, f, document_name: str = None, substrate_source: str = None, parent=None
    ):
        return self.load_der_document(f.read(), document_name, substrate_source, parent)

    def load_b64_document(
        self,
        substrate: str,
        document_name: str = None,
        substrate_source: str = None,
        parent=None,
    ):
        der = base64.b64decode(substrate)

        return self.load_der_document(der, document_name, substrate_source, parent)

    def load_b64_file(
        self, f, document_name: str = None, substrate_source: str = None, parent=None
    ):
        data = f.read()

        if isinstance(data, bytes):
            data = data.decode("us-ascii")

        return self.load_b64_document(data, document_name, substrate_source, parent)

    def load_pem_document(
        self,
        substrate: str,
        document_name: str = None,
        substrate_source: str = None,
        parent=None,
    ):
        m = self._pem_re.match(substrate)

        if m is None:
            raise ValueError("Invalid PEM text")

        b64_text = m.group("pem")

        return self.load_b64_document(b64_text, document_name, substrate_source, parent)

    def load_pem_file(
        self, f, document_name: str = None, substrate_source: str = None, parent=None
    ):
        data = f.read()

        if isinstance(data, bytes):
            data = data.decode("us-ascii")

        return self.load_pem_document(data, document_name, substrate_source, parent)

    @classmethod
    def _is_ascii_armor_start_present(cls, substrate: str):
        first_significant_char = next((c for c in substrate if not c.isspace()), None)

        return first_significant_char == "-"

    def load_document(
        self,
        substrate,
        document_name: str = None,
        substrate_source: str = None,
        parent=None,
    ):
        if isinstance(substrate, bytes):
            try:
                return self.load_der_document(
                    substrate, document_name, substrate_source, parent
                )
            except ValueError:
                substrate = substrate.decode("us-ascii")

        if self._is_ascii_armor_start_present(substrate):
            return self.load_pem_document(
                substrate, document_name, substrate_source, parent
            )
        else:
            return self.load_b64_document(
                substrate, document_name, substrate_source, parent
            )

    def load_file(
        self, f, document_name: str = None, substrate_source: str = None, parent=None
    ):
        substrate = f.read()

        return self.load_document(substrate, document_name, substrate_source, parent)

    def load_document_or_file(
        self,
        substrate,
        document_name: str = None,
        substrate_source: str = None,
        parent=None,
    ):
        try:
            return self.load_file(substrate, document_name, substrate_source, parent)
        except AttributeError:
            return self.load_document(
                substrate, document_name, substrate_source, parent
            )

    def get_file_loader_func(self, document_format: DocumentFormat):
        if document_format == DocumentFormat.BASE64:
            return self.load_b64_file
        elif document_format == DocumentFormat.DER:
            return self.load_der_file
        elif document_format == DocumentFormat.PEM:
            return self.load_pem_file
        elif document_format == DocumentFormat.DETECT:
            return self.load_file
        else:
            raise ValueError(f"Unknown document format: {document_format}")


class RFC5280CertificateDocumentLoader(DocumentLoader):
    def __init__(self):
        super().__init__(RFC5280Certificate, "CERTIFICATE")


class RFC5280CertificateListDocumentLoader(DocumentLoader):
    def __init__(self):
        super().__init__(RFC5280CertificateList, "X509 CRL")


class RFC6960OCSPResponseDocumentLoader(DocumentLoader):
    def __init__(self):
        super().__init__(RFC6960OCSPResponse, "OCSP RESPONSE")


# RFC 5280 Certificate
_RFC5280_CERTIFICATE_LOADER_INSTANCE = RFC5280CertificateDocumentLoader()
load_der_certificate = _RFC5280_CERTIFICATE_LOADER_INSTANCE.load_der_document
load_pem_certificate = _RFC5280_CERTIFICATE_LOADER_INSTANCE.load_pem_document
load_b64_certificate = _RFC5280_CERTIFICATE_LOADER_INSTANCE.load_b64_document
load_certificate = _RFC5280_CERTIFICATE_LOADER_INSTANCE.load_document_or_file
load_der_certificate_file = _RFC5280_CERTIFICATE_LOADER_INSTANCE.load_der_file
load_pem_certificate_file = _RFC5280_CERTIFICATE_LOADER_INSTANCE.load_pem_file
load_b64_certificate_file = _RFC5280_CERTIFICATE_LOADER_INSTANCE.load_b64_file
load_certificate_file = _RFC5280_CERTIFICATE_LOADER_INSTANCE.load_file


# RFC 5280 CRL
_RFC5280_CERTIFICATE_LIST_LOADER_INSTANCE = RFC5280CertificateListDocumentLoader()
load_der_crl = _RFC5280_CERTIFICATE_LIST_LOADER_INSTANCE.load_der_document
load_pem_crl = _RFC5280_CERTIFICATE_LIST_LOADER_INSTANCE.load_pem_document
load_b64_crl = _RFC5280_CERTIFICATE_LIST_LOADER_INSTANCE.load_b64_document
load_crl = _RFC5280_CERTIFICATE_LIST_LOADER_INSTANCE.load_document_or_file
load_der_crl_file = _RFC5280_CERTIFICATE_LIST_LOADER_INSTANCE.load_der_file
load_pem_crl_file = _RFC5280_CERTIFICATE_LIST_LOADER_INSTANCE.load_pem_file
load_b64_crl_file = _RFC5280_CERTIFICATE_LIST_LOADER_INSTANCE.load_b64_file
load_crl_file = _RFC5280_CERTIFICATE_LIST_LOADER_INSTANCE.load_file


# RFC 6960 OCSP Response
_RFC6960_OCSP_RESPONSE_LOADER_INSTANCE = RFC6960OCSPResponseDocumentLoader()
load_der_ocsp_response = _RFC6960_OCSP_RESPONSE_LOADER_INSTANCE.load_der_document
load_pem_ocsp_response = _RFC6960_OCSP_RESPONSE_LOADER_INSTANCE.load_pem_document
load_b64_ocsp_response = _RFC6960_OCSP_RESPONSE_LOADER_INSTANCE.load_b64_document
load_ocsp_response = _RFC6960_OCSP_RESPONSE_LOADER_INSTANCE.load_document_or_file
load_der_ocsp_response_file = _RFC6960_OCSP_RESPONSE_LOADER_INSTANCE.load_der_file
load_pem_ocsp_response_file = _RFC6960_OCSP_RESPONSE_LOADER_INSTANCE.load_pem_file
load_b64_ocsp_response_file = _RFC6960_OCSP_RESPONSE_LOADER_INSTANCE.load_b64_file
load_ocsp_response_file = _RFC6960_OCSP_RESPONSE_LOADER_INSTANCE.load_file



import operator
from typing import Mapping, Tuple, Optional

from dateutil.relativedelta import relativedelta
from pyasn1.type import univ
from pyasn1_alt_modules import rfc8398, rfc5280, rfc4262

import pkilint.adobe.asn1 as adobe_asn1
import pkilint.cabf.cabf_extension
import pkilint.cabf.smime.smime_extension
import pkilint.common
import pkilint.etsi.asn1
import pkilint.pkix.certificate
from pkilint import validation, cabf, document
from pkilint.adobe import adobe_validator
from pkilint.cabf import cabf_extension, cabf_key, cabf_name
from pkilint.cabf.smime import (
    smime_constants,
    smime_name,
    smime_key,
    smime_extension,
    smime_validity,
)
from pkilint.cabf.smime.smime_constants import Generation
from pkilint.common import alternative_name
from pkilint.iso import lei
from pkilint.msft import asn1 as microsoft_asn1
from pkilint.msft import msft_name
from pkilint.pkix import certificate, time
from pkilint.pkix.certificate import certificate_validity
from pkilint.pkix.general_name import OTHER_NAME_MAPPINGS as PKIX_OTHERNAME_MAPPINGS

OTHER_NAME_MAPPINGS = {
    **PKIX_OTHERNAME_MAPPINGS,
    rfc8398.id_on_SmtpUTF8Mailbox: rfc8398.SmtpUTF8Mailbox(),
    microsoft_asn1.id_on_UserPrincipalName: microsoft_asn1.UserPrincipalName(),
}


def determine_validation_level_and_generation(
    cert: certificate.RFC5280Certificate,
    config: Mapping[
        univ.ObjectIdentifier,
        Tuple[smime_constants.ValidationLevel, smime_extension.Generation],
    ] = None,
):
    oids = cert.policy_oids

    for v in smime_constants.ValidationLevel:
        for g in smime_constants.Generation:
            oid = smime_constants.get_policy_oid(v, g)

            if oid in oids:
                return v, g

    if config is not None:
        for o in oids:
            v_g = config.get(o)

            if v_g is not None:
                return v_g

    return None


def _has_subject_attr(cert, attr):
    return any(cert.get_subject_attributes_by_type(attr))


def _get_first_subject_attr_dirstring_value(cert, attr, attr_asn1_cls):
    attrs = cert.get_subject_attributes_by_type(attr)

    if any(attrs):
        attr, _ = attrs[0]

        attr_value_pdu = attr.children["value"].pdu

        decoded_value = document.decode_substrate(cert, attr_value_pdu, attr_asn1_cls())

        # assume DirectoryString
        _, attr_value_choice_value = decoded_value.child

        return str(attr_value_choice_value.pdu)
    else:
        return None


def guess_validation_level_and_generation(
    cert: certificate.RFC5280Certificate,
    config: Mapping[
        univ.ObjectIdentifier,
        Tuple[smime_constants.ValidationLevel, smime_extension.Generation],
    ] = None,
):
    v_g = determine_validation_level_and_generation(cert, config)

    if v_g is not None:
        return v_g

    # assume Legacy generation
    g = smime_constants.Generation.LEGACY

    o = _get_first_subject_attr_dirstring_value(
        cert, rfc5280.id_at_organizationName, rfc5280.X520OrganizationName
    )
    has_o = o is not None
    cn = _get_first_subject_attr_dirstring_value(
        cert, rfc5280.id_at_commonName, rfc5280.X520CommonName
    )
    has_cn = cn is not None
    has_natural_name = _has_subject_attr(
        cert, rfc5280.id_at_surname
    ) or _has_subject_attr(cert, rfc5280.id_at_givenName)

    if has_o and (has_natural_name or (has_cn and o != cn and "@" not in cn)):
        v = smime_constants.ValidationLevel.SPONSORED
    elif has_o:
        v = smime_constants.ValidationLevel.ORGANIZATION
    elif has_natural_name:
        v = smime_constants.ValidationLevel.INDIVIDUAL
    else:
        v = smime_constants.ValidationLevel.MAILBOX

    return v, g


_SMIME_EXTENSION_MAPPINGS = {
    **cabf.EXTENSION_MAPPINGS,
    **lei.EXTENSION_MAPPINGS,
    **rfc4262.certificateExtensionsMap,
    **adobe_asn1.EXTENSION_MAPPINGS,
}


def create_decoding_validators():
    return pkilint.pkix.certificate.create_decoding_validators(
        cabf.NAME_ATTRIBUTE_MAPPINGS,
        _SMIME_EXTENSION_MAPPINGS,
        [
            certificate.create_other_name_decoder(OTHER_NAME_MAPPINGS),
            certificate.create_qc_statements_decoder(
                pkilint.etsi.asn1.ETSI_QC_STATEMENTS_MAPPINGS
            ),
        ],
    )


def create_spki_validation_container():
    return certificate.create_spki_validator_container(
        [
            smime_key.SmimeAllowedPublicKeyAlgorithmEncodingValidator(
                path="certificate.tbsCertificate.subjectPublicKeyInfo.algorithm"
            ),
            cabf_key.RsaKeyValidator(),
            cabf_key.EcdsaKeyValidator(),
            smime_key.GmailAllowedModulusLengthValidator(),
        ]
    )


def create_extensions_validator_container(validation_level, generation):
    return certificate.create_extensions_validator_container(
        [
            smime_extension.RequiredPolicyIdentifierValidator(
                validation_level, generation
            ),
            smime_extension.CertificatePoliciesPresenceValidator(),
            smime_extension.ExtendedKeyUsagePresenceValidator(),
            smime_extension.CabfSmimeKeyUsagePresenceValidator(),
            cabf_extension.AuthorityInformationAccessPresenceValidator(
                validation.ValidationFindingSeverity.WARNING
            ),
            smime_extension.CrlDistributionPointPresenceValidator(),
            smime_extension.SubjectAlternativeNamePresenceValidator(),
            smime_extension.AllowedExtendedKeyUsageValidator(generation),
            smime_extension.AllowedKeyUsageValidator(generation),
            smime_extension.EndEntityValidator(),
            cabf_extension.CpsUriHttpValidator(),
            cabf_extension.AuthorityInformationAccessContainsHttpUriValidator(),
            cabf_extension.CrlDpContainsHttpUriValidator(),
            smime_extension.SubjectAlternativeNameContainsEmailAddressValidator(),
            smime_extension.SubjectAlternativeNameProhibitedGeneralNameTypesValidator(
                generation
            ),
            smime_extension.CabfSmimeKeyUsageCriticalityValidator(),
            smime_extension.GmailAuthorityInfoAccessCaIssuersValidator(),
            msft_name.UserPrincipalNameSyntaxValidator(),
            smime_extension.AllowedCrldpFullNameValidator(generation),
            smime_extension.SmimeUserNoticeValidator(),
            smime_extension.AllowedAiaUriSchemeValidator(generation),
            smime_extension.LeiCriticalityValidator(),
            smime_extension.LeiRoleCriticalityValidator(),
            smime_extension.LeiPresenceValidator(validation_level),
            smime_extension.LeiRolePresenceValidator(validation_level),
            lei.LeiExtensionValueSyntaxValidator(),
            cabf_extension.CertificatePoliciesCriticalityValidator(),
            cabf_extension.CabfCrlDpCriticalityValidator(),
            cabf_extension.CabfAuthorityKeyIdentifierValidator(),
            smime_extension.SubjectDirectoryAttributesPresenceValidator(
                validation_level, generation
            ),
            smime_extension.QCStatementsCriticalityValidator(),
            alternative_name.create_internal_name_validator_container(
                cabf_name.VALIDATION_INTERNAL_DOMAIN_NAME,
                cabf_name.VALIDATION_INTERNAL_IP_ADDRESS,
                allow_onion_tld=False,
            ),
            alternative_name.create_cpsuri_internal_domain_name_validator(
                cabf_name.VALIDATION_INTERNAL_DOMAIN_NAME
            ),
            adobe_validator.AdobeTimestampValidator(),
            smime_extension.AdobeTimestampCriticalityValidator(),
            smime_extension.AdobeTimestampPresenceValidator(generation),
            smime_extension.AdobeArchiveRevInfoCriticalityValidator(),
            smime_extension.AdobeArchiveRevInfoPresenceValidator(generation),
        ]
    )


def create_validity_validators(
    generation, validity_period_start_retriever: document.ValidityPeriodStartRetriever
):
    days = 1185 if generation == Generation.LEGACY else 825

    threshold_error = (
        operator.le,
        relativedelta(days=days),
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            f"cabf.smime.certificate_validity_period_exceeds_{days}_days",
        ),
    )

    threshold_warning = (
        operator.le,
        relativedelta(days=days - 1, hours=23, minutes=59, seconds=59),
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.WARNING,
            "cabf.smime.certificate_validity_period_at_maximum",
        ),
    )

    validators = [
        time.ValidityPeriodThresholdsValidator(
            path="certificate.tbsCertificate.validity.notBefore",
            end_validity_node_retriever=lambda n: n.navigate("^.notAfter"),
            inclusive_second=True,
            validity_period_thresholds=[threshold_error, threshold_warning],
        )
    ]

    if generation == smime_constants.Generation.LEGACY:
        validators.append(
            smime_validity.LegacyGenerationSunsetValidator(
                validity_period_start_retriever
            )
        )

    return validators


def create_subscriber_validators(
    validation_level,
    generation,
    validity_period_start_retriever: Optional[
        document.ValidityPeriodStartRetriever
    ] = None,
):
    if validity_period_start_retriever is None:
        validity_period_start_retriever = (
            certificate_validity.CertificateValidityPeriodStartRetriever()
        )

    return [
        smime_name.create_subscriber_certificate_subject_validator_container(
            validation_level, generation
        ),
        create_spki_validation_container(),
        certificate.create_issuer_validator_container([]),
        certificate.create_validity_validator_container(
            create_validity_validators(generation, validity_period_start_retriever)
        ),
        create_extensions_validator_container(validation_level, generation),
        smime_key.SmimeAllowedSignatureAlgorithmEncodingValidator(
            path="certificate.tbsCertificate.signature"
        ),
        cabf_extension.CabfExtensionsPresenceValidator(),
    ]



from pyasn1_alt_modules import rfc5280

from pkilint import validation
from pkilint.pkix import general_name


class SubjectEmailAddressInSanValidator(validation.Validator):
    VALIDATION_SUBJECT_EMAIL_NOT_IN_SAN = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.subject_email_address_not_in_san",
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.EmailAddress,
            validations=[self.VALIDATION_SUBJECT_EMAIL_NOT_IN_SAN],
        )

    def validate(self, node):
        san_ext_idx = node.document.get_extension_by_oid(rfc5280.id_ce_subjectAltName)

        if san_ext_idx is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SUBJECT_EMAIL_NOT_IN_SAN,
                "Certificate does not have SAN extension",
            )

        ext, _ = san_ext_idx

        email_address = str(node.pdu)

        if not any(
            str(rfc822name_node.pdu) == email_address
            for rfc822name_node in node.document.get_san_general_names_by_type(
                general_name.GeneralNameTypeName.RFC822_NAME
            )
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_SUBJECT_EMAIL_NOT_IN_SAN,
                f'Subject DN e-mail address "{email_address}" not found in SAN',
            )



from pyasn1_alt_modules import rfc5280, rfc6960, rfc6962

import pkilint.common
from pkilint import validation, common
from pkilint.itu import bitstring
from pkilint.pkix import Rfc2119Word
from pkilint.pkix.certificate.certificate_extension import KeyUsageBitName

_CODE_CLASSIFIER = "cabf.serverauth.ocsp_responder"


class OcspResponderKeyUsageValidator(validation.Validator):
    """Validates that the content of the key usage extension conforms with BR 7.1.2.8.7."""

    VALIDATION_DIGSIG_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        _CODE_CLASSIFIER + ".digitalsignature_bit_missing",
    )

    VALIDATION_PROHIBITED_KU_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        _CODE_CLASSIFIER + ".prohibited_ku_present",
    )

    _PROHIBITED_KUS = {str(n) for n in rfc5280.KeyUsage.namedValues} - {
        KeyUsageBitName.DIGITAL_SIGNATURE
    }

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_DIGSIG_MISSING,
                self.VALIDATION_PROHIBITED_KU_PRESENT,
            ],
            pdu_class=rfc5280.KeyUsage,
        )

    def validate(self, node):
        if not bitstring.has_named_bit(node, KeyUsageBitName.DIGITAL_SIGNATURE):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_DIGSIG_MISSING
            )

        prohibited_kus_asserted = sorted(
            (k for k in self._PROHIBITED_KUS if bitstring.has_named_bit(node, k))
        )

        if any(prohibited_kus_asserted):
            prohibited_kus_str = ", ".join(prohibited_kus_asserted)

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_KU_PRESENT,
                f"Prohibited KUs present: {prohibited_kus_str}",
            )


class OcspAuthorityInformationAccessAccessMethodPresenceValidator(
    common.AuthorityInformationAccessAccessMethodPresenceValidator
):
    """Validates that the content of the AIA extension conforms with BR 7.1.2.8.3."""

    _ACCESS_METHOD_ALLOWANCES = {
        rfc5280.id_ad_ocsp: Rfc2119Word.SHOULD_NOT,
    }

    def __init__(self):
        super().__init__(
            self._ACCESS_METHOD_ALLOWANCES, _CODE_CLASSIFIER, Rfc2119Word.MUST_NOT
        )


class OcspExtensionAllowanceValidator(
    pkilint.common.ExtensionIdentifierAllowanceValidator
):
    """Validates that the included extensions conform with BR 7.1.2.8.2."""

    _EXTENSION_ALLOWANCES = {
        rfc5280.id_ce_authorityKeyIdentifier: Rfc2119Word.MUST,
        rfc5280.id_ce_extKeyUsage: Rfc2119Word.MUST,
        rfc6960.id_pkix_ocsp_nocheck: Rfc2119Word.MUST,
        rfc5280.id_ce_keyUsage: Rfc2119Word.MUST,
        rfc5280.id_ce_basicConstraints: Rfc2119Word.MAY,
        rfc5280.id_ce_nameConstraints: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_subjectAltName: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_subjectKeyIdentifier: Rfc2119Word.SHOULD,
        rfc5280.id_pe_authorityInfoAccess: Rfc2119Word.SHOULD_NOT,
        rfc5280.id_ce_certificatePolicies: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_cRLDistributionPoints: Rfc2119Word.MUST_NOT,
        rfc6962.id_ce_embeddedSCT: Rfc2119Word.MAY,
    }

    def __init__(self):
        super().__init__(
            self._EXTENSION_ALLOWANCES, _CODE_CLASSIFIER, Rfc2119Word.SHOULD_NOT
        )


class OcspEkuAllowanceValidator(pkilint.common.ExtendedKeyUsageAllowanceValidator):
    """Validates that the extended key usage value conforms to BR 7.1.2.8.5."""

    _EKU_ALLOWANCES = {
        rfc5280.id_kp_OCSPSigning: Rfc2119Word.MUST,
    }

    def __init__(self):
        super().__init__(self._EKU_ALLOWANCES, _CODE_CLASSIFIER, Rfc2119Word.MUST_NOT)


class OcspBasicConstraintsValidator(validation.Validator):
    """Validates that the basic constraints extension value conforms to BR 7.1.2.8.4."""

    VALIDATION_CA_BIT_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        _CODE_CLASSIFIER + ".basic_constraints_ca_bit_set",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_CA_BIT_SET, pdu_class=rfc5280.BasicConstraints
        )

    def validate(self, node):
        if bool(node.children["cA"].pdu):
            raise validation.ValidationFindingEncountered(self.VALIDATION_CA_BIT_SET)



from pkilint import pkix

from pkilint.pkix import crl, extension, name
from pkilint.rest import model


def create_crl_linter(
    validity_additional_validators=None, doc_additional_validators=None
):
    if doc_additional_validators is None:
        doc_additional_validators = []
    if validity_additional_validators is None:
        validity_additional_validators = []

    return model.Linter(
        validator=crl.create_pkix_crl_validator_container(
            [
                pkix.create_attribute_decoder(name.ATTRIBUTE_TYPE_MAPPINGS),
                pkix.create_extension_decoder(extension.EXTENSION_MAPPINGS),
            ],
            [
                crl.create_issuer_validator_container([]),
                crl.create_validity_validator_container(validity_additional_validators),
                crl.create_extensions_validator_container([]),
            ]
            + doc_additional_validators,
        ),
        name="crl_linter",
    )



from pyasn1_alt_modules import rfc5280

from pkilint import common
from pkilint import validation
from pkilint.common import organization_id
from pkilint.etsi import etsi_shared
from pkilint.itu import x520_name, asn1_util
from pkilint.pkix import Rfc2119Word, name

_REQUIRED_ATTRIBUTES = {
    rfc5280.id_at_countryName,
    rfc5280.id_at_organizationName,
    x520_name.id_at_organizationIdentifier,
    rfc5280.id_at_commonName,
}


class LegalPersonSubjectAttributeAllowanceValidator(
    common.AttributeIdentifierAllowanceValidator
):
    """
    LEG-4.2.1-2: The subject field shall include at least the following attributes as specified in Recommendation
    ITU-T X.520
    """

    _CODE_CLASSIFIER = "etsi.en_319_412_3.leg-4.2.1-2"

    _ATTRIBUTE_ALLOWANCES = {a: Rfc2119Word.MUST for a in _REQUIRED_ATTRIBUTES}

    def __init__(self):
        super().__init__(
            self._ATTRIBUTE_ALLOWANCES, self._CODE_CLASSIFIER, Rfc2119Word.MAY
        )


class LegalPersonDuplicateAttributeAllowanceValidator(validation.Validator):
    """
    LEG-4.2.1-3: Only one instance of each of these attributes shall be present.
    """

    VALIDATION_PROHIBITED_DUPLICATE_ATTRIBUTE_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_3.leg-4.2.1-3.prohibited_duplicate_attribute_present",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_PROHIBITED_DUPLICATE_ATTRIBUTE_PRESENT],
            pdu_class=rfc5280.Name,
        )

    def validate(self, node):
        attr_counts = name.get_name_attribute_counts(node)

        for a in _REQUIRED_ATTRIBUTES:
            if attr_counts[a] > 1:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_PROHIBITED_DUPLICATE_ATTRIBUTE_PRESENT,
                    f"Prohibited duplicate attribute present: {a}",
                )


class LegalPersonOrganizationAttributesEqualityValidator(validation.Validator):
    """
    LEG-4.2.1-6: The organizationIdentifier attribute shall contain an identification of the subject organization
    different from the organization name.
    """

    VALIDATION_ORGID_ORGNAME_ATTRIBUTE_VALUES_EQUAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_3.leg-4.2.1-6.organization_id_and_organization_name_attribute_values_equal",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_ORGID_ORGNAME_ATTRIBUTE_VALUES_EQUAL],
            pdu_class=rfc5280.Name,
        )

    def validate(self, node):
        # only get the first instance of the attributes
        orgname_attr_and_idx = next(
            iter(
                name.get_name_attributes_by_type(node, rfc5280.id_at_organizationName)
            ),
            None,
        )
        orgid_attr_and_idx = next(
            iter(
                name.get_name_attributes_by_type(
                    node, x520_name.id_at_organizationIdentifier
                )
            ),
            None,
        )

        if orgname_attr_and_idx and orgid_attr_and_idx:
            orgname_attr, _ = orgname_attr_and_idx
            orgid_attr, _ = orgid_attr_and_idx

            orgname = asn1_util.get_string_value_from_attribute_node(orgname_attr)
            orgid = asn1_util.get_string_value_from_attribute_node(orgid_attr)

            # if any of the attributes were not decoded, then return early
            if orgname is None or orgid is None:
                return

            if orgname.casefold() == orgid.casefold():
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_ORGID_ORGNAME_ATTRIBUTE_VALUES_EQUAL,
                    f'Organization name and identifier attribute values are equal: "{orgname}"',
                )


class LegalPersonKeyUsageValidator(etsi_shared.KeyUsageValidator):
    """
    LEG-4.3.1-3: Certificates used to validate commitment to signed content (e.g. documents, agreements and/or
    transactions) shall be limited to type A, B or F.
    """

    VALIDATION_INVALID_CONTENT_COMMITMENT_SETTING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_3.leg-4.3.1-3.invalid_content_commitment_setting",
    )

    """
    LEG-4.3.1-4: Of these alternatives, type A should be used (see the security note 2 below).
    """
    VALIDATION_NON_PREFERRED_CONTENT_COMMITMENT_SETTING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "etsi.en_319_412_3.leg-4.3.1-4.non_preferred_content_commitment_setting",
    )

    def __init__(self, is_content_commitment_type):
        super().__init__(
            is_content_commitment_type,
            self.VALIDATION_INVALID_CONTENT_COMMITMENT_SETTING,
            self.VALIDATION_NON_PREFERRED_CONTENT_COMMITMENT_SETTING,
        )


class LegalPersonCountryCodeValidator(validation.Validator):
    """
    LEG-4.2.1-4: The countryName attribute shall specify the country in which the subject (legal person) is established.
    """

    VALIDATION_UNKNOWN_COUNTRY_CODE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "etsi.en_319_412_3.leg-4.2.1-4.unknown_country_code",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_UNKNOWN_COUNTRY_CODE],
            pdu_class=rfc5280.X520countryName,
        )

    def validate(self, node):
        value_str = str(node.pdu)

        if value_str not in organization_id.ISO3166_1_COUNTRY_CODES:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNKNOWN_COUNTRY_CODE,
                f'Unknown country code: "{value_str}"',
            )



from cryptography.hazmat.primitives import hashes


def calculate_hash(octets: bytes, hash_algo: hashes.HashAlgorithm) -> bytes:
    h = hashes.Hash(hash_algo)
    h.update(octets)

    return h.finalize()


def calculate_sha1_hash(octets: bytes) -> bytes:
    return calculate_hash(octets, hashes.SHA1())



def has_named_bit(node, bit_name):
    bit = node.pdu.namedValues[bit_name]
    return len(node.pdu) > bit and node.pdu[bit] != 0


def get_asserted_bit_set(node):
    return {str(b) for b in node.pdu.namedValues if has_named_bit(node, str(b))}



import enum

from pyasn1_alt_modules import rfc5280

from pkilint import validation, pkix, document
from pkilint.document import Document
from pkilint.pkix import name, extension, time, general_name
from pkilint.pkix.crl import crl_validator, crl_extension, crl_validity


@enum.unique
class CertificateRevocationListType(enum.IntEnum):
    CRL = 0
    ARL = 1


class RFC5280CertificateList(Document):
    def __init__(self, substrate_source, substrate, name=None, parent=None):
        super().__init__(
            rfc5280.CertificateList(), substrate_source, substrate, name, parent
        )

    @property
    def this_update(self):
        try:
            return time.parse_time_node(self.root.navigate("tbsCertList.thisUpdate"))
        except ValueError:
            return pkix.MAXIMUM_TIME_DATETIME

    @property
    def next_update(self):
        try:
            return time.parse_time_node(self.root.navigate("tbsCertList.nextUpdate"))
        except document.PDUNavigationFailedError:
            return None
        except ValueError:
            return pkix.MAXIMUM_TIME_DATETIME

    def get_extension_by_oid(self, oid):
        tbs_crl = self.root.children["tbsCertList"]

        # ensure there's extensions
        if "crlExtensions" not in tbs_crl.children:
            return None

        extensions_node = tbs_crl.children["crlExtensions"]

        for ext_idx, extension_node in extensions_node.children.items():
            ext_oid = extension_node.children["extnID"].pdu

            if ext_oid == oid:
                return extension_node, int(ext_idx)

        return None


def create_issuer_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    additional_validators.append(name.EmptyNameValidator())

    return pkix.create_name_validator_container(
        additional_validators, path="certificateList.tbsCertList.issuer"
    )


def create_validity_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    return validation.ValidatorContainer(
        validators=[
            time.TimeCorrectEncodingValidator(),
        ]
        + additional_validators,
        pdu_class=rfc5280.Time,
    )


def create_extensions_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []
    return validation.ValidatorContainer(
        validators=[
            # extension.PermittedExtensionValidator(
            #    known_oids=known_oids
            # ),
            extension.UniqueExtensionValidator(),
            crl_extension.CrlNumberCriticalityValidator(),
            extension.AuthorityKeyIdentifierValidator(),
        ]
        + additional_validators,
        path="certificateList.tbsCertList.crlExtensions",
    )


def create_pkix_crl_validator_container(decoding_validators, validators):
    decoding_validator_container = [
        validation.ValidatorContainer(
            validators=decoding_validators, path="certificateList"
        )
    ]

    validators += [
        crl_validator.VersionPresenceValidator(),
        crl_validator.CorrectVersionValidator(),
        crl_extension.CrlNumberPresenceValidator(),
        crl_extension.AuthorityKeyIdentifierPresenceValidator(),
        crl_validator.SignatureAlgorithmMatchValidator(),
        crl_validator.RevokedCertificatesEmptyValidator(),
        crl_extension.CrlReasonCodeCriticalityValidator(),
        time.UtcTimeCorrectSyntaxValidator(),
        time.GeneralizedTimeCorrectSyntaxValidator(),
        crl_validity.CrlSaneValidityPeriodValidator(),
        pkix.CertificateSerialNumberValidator(),
        crl_extension.CrlNumberValueValidator(),
        general_name.GeneralNameValidatorContainer(),
    ]

    return validation.ValidatorContainer(
        validators=decoding_validator_container + validators
    )



from pyasn1.type import univ, namedtype, namedval
from pyasn1_alt_modules import rfc5280


_ADOBE_X509_OID_ARC = univ.ObjectIdentifier("1.2.840.113583.1.1.9")


id_adobe_timestamp = univ.ObjectIdentifier(_ADOBE_X509_OID_ARC.asTuple() + (1,))


class AdobeExtensionVersion(univ.Integer):
    pass


AdobeExtensionVersion.componentType = namedval.NamedValues(
    ("v1", 1),
)


class AdobeTimestamp(univ.Sequence):
    pass


AdobeTimestamp.componentType = namedtype.NamedTypes(
    namedtype.NamedType("version", AdobeExtensionVersion()),
    namedtype.NamedType("location", rfc5280.GeneralName()),
    namedtype.DefaultedNamedType("requiresAuth", univ.Boolean().subtype(value=False)),
)


id_adobe_archiverevinfo = univ.ObjectIdentifier(_ADOBE_X509_OID_ARC.asTuple() + (2,))


class AdobeArchiveRevInfo(univ.Sequence):
    pass


AdobeArchiveRevInfo.componentType = namedtype.NamedTypes(
    namedtype.NamedType("version", AdobeExtensionVersion())
)


EXTENSION_MAPPINGS = {
    id_adobe_timestamp: AdobeTimestamp(),
    id_adobe_archiverevinfo: AdobeArchiveRevInfo(),
}



from pyasn1.type import univ
from pyasn1.type.constraint import ValueSizeConstraint

from pkilint import document
from pkilint.nist.asn1 import csor


ML_DSA_44_PublicKeySize = 1312
ML_DSA_65_PublicKeySize = 1952
ML_DSA_87_PublicKeySize = 2592


class MlDsa44PublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(ML_DSA_44_PublicKeySize, ML_DSA_44_PublicKeySize)


class MlDsa65PublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(ML_DSA_65_PublicKeySize, ML_DSA_65_PublicKeySize)


class MlDsa87PublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(ML_DSA_87_PublicKeySize, ML_DSA_87_PublicKeySize)


ALGORITHM_OID_TO_KEY_MAPPINGS = {
    # pure
    csor.id_ml_dsa_44: MlDsa44PublicKey(),
    csor.id_ml_dsa_65: MlDsa65PublicKey(),
    csor.id_ml_dsa_87: MlDsa87PublicKey(),
    # pre-hashed
    csor.id_hash_ml_dsa_44_with_sha512: MlDsa44PublicKey(),
    csor.id_hash_ml_dsa_65_with_sha512: MlDsa65PublicKey(),
    csor.id_hash_ml_dsa_87_with_sha512: MlDsa87PublicKey(),
}

ALGORITHM_OID_TO_PARAMETER_MAPPINGS = {
    k: document.ValueDecoder.VALUE_NODE_ABSENT
    for k in ALGORITHM_OID_TO_KEY_MAPPINGS.keys()
}



from cryptography import exceptions
from cryptography.hazmat.primitives.asymmetric import (
    ec,
    rsa,
    padding,
    dsa,
    ed25519,
    ed448,
)
from pyasn1.type import univ
from pyasn1_alt_modules import rfc3279, rfc5480, rfc8410

from pkilint import document
from pkilint.document import PDUNode
from pkilint.nist.asn1 import fips_203, fips_204, fips_205

SUBJECT_PUBLIC_KEY_ALGORITHM_IDENTIFIER_MAPPINGS = {
    rfc3279.rsaEncryption: rfc5480.RSAPublicKey(),
    rfc5480.id_ecPublicKey: rfc5480.ECPoint(),
    rfc5480.id_ecDH: rfc5480.ECPoint(),
    rfc5480.id_ecMQV: rfc5480.ECPoint(),
    rfc8410.id_Ed448: univ.OctetString(),
    rfc8410.id_Ed25519: univ.OctetString(),
    rfc8410.id_X448: univ.OctetString(),
    rfc8410.id_X25519: univ.OctetString(),
    **fips_203.ALGORITHM_OID_TO_KEY_MAPPINGS,
    **fips_204.ALGORITHM_OID_TO_KEY_MAPPINGS,
    **fips_205.ALGORITHM_OID_TO_KEY_MAPPINGS,
}

SUBJECT_KEY_PARAMETER_ALGORITHM_IDENTIFIER_MAPPINGS = {
    rfc3279.rsaEncryption: univ.Null(),
    rfc5480.id_ecPublicKey: rfc5480.ECParameters(),
    rfc5480.id_ecDH: rfc5480.ECParameters(),
    rfc5480.id_ecMQV: rfc5480.ECParameters(),
    **{
        o: document.ValueDecoder.VALUE_NODE_ABSENT
        for o in (
            rfc8410.id_Ed448,
            rfc8410.id_Ed25519,
            rfc8410.id_X448,
            rfc8410.id_X25519,
        )
    },
    **fips_203.ALGORITHM_OID_TO_PARAMETER_MAPPINGS,
    **fips_204.ALGORITHM_OID_TO_PARAMETER_MAPPINGS,
    **fips_205.ALGORITHM_OID_TO_PARAMETER_MAPPINGS,
}

EC_CURVE_OID_TO_OBJECT_MAPPINGS = {
    rfc5480.secp256r1: ec.SECP256R1(),
    rfc5480.secp384r1: ec.SECP384R1(),
    rfc5480.secp521r1: ec.SECP521R1(),
}

_EC_PUBLIC_KEY_OIDS = {rfc5480.id_ecPublicKey, rfc5480.id_ecDH, rfc5480.id_ecMQV}


def convert_spki_to_object(spki_node: PDUNode):
    key_type = spki_node.navigate("algorithm.algorithm").pdu

    if key_type == rfc3279.rsaEncryption:
        rsa_public_key = spki_node.navigate("subjectPublicKey.rSAPublicKey")

        modulus = rsa_public_key.children["modulus"].pdu
        exponent = rsa_public_key.children["publicExponent"].pdu

        return rsa.RSAPublicNumbers(int(exponent), int(modulus)).public_key()
    elif key_type in _EC_PUBLIC_KEY_OIDS:
        curve_oid = spki_node.navigate(
            "algorithm.parameters.eCParameters.namedCurve"
        ).pdu

        curve = EC_CURVE_OID_TO_OBJECT_MAPPINGS.get(curve_oid)
        if curve is not None:
            return ec.EllipticCurvePublicKey.from_encoded_point(
                curve, spki_node.navigate("subjectPublicKey").pdu.asOctets()
            )
    elif key_type == rfc8410.id_Ed448:
        return ed448.Ed448PublicKey.from_public_bytes(
            spki_node.navigate("subjectPublicKey").pdu.asOctets()
        )
    elif key_type == rfc8410.id_Ed25519:
        return ed25519.Ed25519PublicKey.from_public_bytes(
            spki_node.navigate("subjectPublicKey").pdu.asOctets()
        )

    # TODO: others
    return None


def verify_signature(public_key, message, signature, signature_hash_algorithm=None):
    try:
        # TODO: add support for RSASSA-PSS

        if isinstance(public_key, rsa.RSAPublicKey):
            public_key.verify(
                signature, message, padding.PKCS1v15(), signature_hash_algorithm
            )
        elif isinstance(public_key, dsa.DSAPublicKey):
            public_key.verify(signature, message, signature_hash_algorithm)
        elif isinstance(public_key, ec.EllipticCurvePublicKey):
            public_key.verify(signature, message, ec.ECDSA(signature_hash_algorithm))
        elif isinstance(public_key, ed25519.Ed25519PublicKey):
            public_key.verify(signature, message)
        elif isinstance(public_key, ed448.Ed448PublicKey):
            public_key.verify(signature, message)
        else:
            type_name = type(public_key).__name__

            raise exceptions.UnsupportedAlgorithm(
                f'Unsupported public key type "{type_name}"'
            )
    except exceptions.InvalidSignature:
        return False

    return True



from typing import Callable, Mapping, Optional

from pyasn1.type import univ
from pyasn1_alt_modules import rfc5280, rfc6962, rfc6960, rfc3739

from pkilint import validation, document, pkix
from pkilint.cabf.asn1 import ev_guidelines
from pkilint.itu import x520_name
from pkilint.pkix import Rfc2119Word

OID_TO_CODE_NAME = {
    # EKUs
    rfc5280.id_kp_codeSigning: "codesigning",
    rfc5280.id_kp_emailProtection: "emailprotection",
    rfc5280.id_kp_timeStamping: "timestamping",
    rfc5280.id_kp_OCSPSigning: "ocspsigning",
    rfc5280.anyExtendedKeyUsage: "anyeku",
    rfc6962.id_kp_precertificateSigning: "precertsigning",
    rfc5280.id_kp_serverAuth: "serverauth",
    rfc5280.id_kp_clientAuth: "clientauth",
    # extensions
    rfc5280.id_ce_authorityKeyIdentifier: "authority_key_identifier",
    rfc5280.id_ce_basicConstraints: "basic_constraints",
    rfc5280.id_ce_certificatePolicies: "certificate_policies",
    rfc5280.id_ce_cRLDistributionPoints: "crl_distribution_points",
    rfc5280.id_ce_keyUsage: "key_usage",
    rfc5280.id_ce_subjectKeyIdentifier: "subject_key_identifier",
    rfc5280.id_ce_extKeyUsage: "extended_key_usage",
    rfc5280.id_pe_authorityInfoAccess: "authority_info_access",
    rfc5280.id_ce_nameConstraints: "name_constraints",
    rfc6962.id_ce_embeddedSCT: "sct_list",
    rfc5280.id_ce_subjectAltName: "subject_altname",
    rfc6962.id_ce_criticalPoison: "precert_poison",
    rfc6960.id_pkix_ocsp_nocheck: "ocsp_nocheck",
    rfc3739.id_pe_qcStatements: "qc_statements",
    rfc5280.id_ce_policyMappings: "policy_mappings",
    rfc5280.id_ce_policyConstraints: "policy_constraints",
    rfc5280.id_ce_inhibitAnyPolicy: "inhibit_any_policy",
    # AIA access methods
    rfc5280.id_ad_ocsp: "ocsp",
    rfc5280.id_ad_caIssuers: "ca_issuers",
    # attributes
    rfc5280.id_at_countryName: "country",
    rfc5280.id_at_stateOrProvinceName: "state_or_province",
    rfc5280.id_at_localityName: "locality",
    x520_name.id_at_postalCode: "postal_code",
    x520_name.id_at_streetAddress: "street_address",
    rfc5280.id_at_organizationName: "organization_name",
    rfc5280.id_at_commonName: "common_name",
    rfc5280.id_at_organizationalUnitName: "organizational_unit_name",
    rfc5280.id_at_serialNumber: "serial_number",
    x520_name.id_at_businessCategory: "business_category",
    ev_guidelines.id_evat_jurisdiction_countryName: "jurisdiction_country",
    ev_guidelines.id_evat_jurisdiction_stateOrProvinceName: "jurisdiction_state_or_province",
    ev_guidelines.id_evat_jurisdiction_localityName: "jurisdiction_locality",
    rfc5280.id_at_surname: "surname",
    rfc5280.id_at_givenName: "given_name",
    x520_name.id_at_organizationIdentifier: "organization_identifier",
}


class ElementIdentifierAllowanceValidator(validation.Validator):
    # use global mappings by default
    _OID_TO_CODE_NAME = OID_TO_CODE_NAME

    @classmethod
    def _create_finding(
        cls, fmt: str, rfc2119word: pkix.Rfc2119Word, o: univ.ObjectIdentifier
    ):
        if rfc2119word == pkix.Rfc2119Word.MAY:
            return None
        else:
            return validation.ValidationFinding(
                rfc2119word.to_severity, fmt.format(oid=cls._OID_TO_CODE_NAME[o])
            )

    def __init__(
        self,
        element_name: str,
        element_oid_retriever: Callable[[document.PDUNode], document.PDUNode],
        known_element_allowances: Mapping[univ.ObjectIdentifier, pkix.Rfc2119Word],
        unexpected_presence_code_format: str = None,
        unexpected_absence_code_format: str = None,
        unknown_element_presence_finding: Optional[validation.ValidationFinding] = None,
        **kwargs,
    ):
        self._element_name = element_name
        self._element_oid_retriever = element_oid_retriever

        self._expected_element_presences = {
            o: self._create_finding(unexpected_absence_code_format, w, o)
            for o, w in known_element_allowances.items()
            if w in {Rfc2119Word.MAY, Rfc2119Word.SHOULD, Rfc2119Word.MUST}
        }
        self._expected_element_absences = {
            o: self._create_finding(unexpected_presence_code_format, w, o)
            for o, w in known_element_allowances.items()
            if w in {Rfc2119Word.MAY, Rfc2119Word.SHOULD_NOT, Rfc2119Word.MUST_NOT}
        }

        self._unknown_element_presence_finding = unknown_element_presence_finding

        validations = [
            a for a in self._expected_element_presences.values() if a is not None
        ] + [a for a in self._expected_element_absences.values() if a is not None]

        if unknown_element_presence_finding is not None:
            validations.append(unknown_element_presence_finding)

        super().__init__(validations=list(validations), **kwargs)

    def validate(self, node):
        oids = {self._element_oid_retriever(n).pdu for n in node.children.values()}

        # process unexpected present elements
        finding_descriptions = [
            validation.ValidationFindingDescription(v, None)
            for o, v in self._expected_element_absences.items()
            if o in oids and v is not None
        ]

        # process unexpected absent elements
        finding_descriptions.extend(
            [
                validation.ValidationFindingDescription(v, None)
                for o, v in self._expected_element_presences.items()
                if o not in oids and v is not None
            ]
        )

        # process unknown elements
        if self._unknown_element_presence_finding is not None:
            finding_descriptions.extend(
                [
                    validation.ValidationFindingDescription(
                        self._unknown_element_presence_finding,
                        f"Unknown {self._element_name} present: {str(o)}",
                    )
                    for o in oids
                    if o not in self._expected_element_presences
                    and o not in self._expected_element_absences
                ]
            )

        return validation.ValidationResult(self, node, finding_descriptions)


class ExtensionsPresenceValidator(validation.Validator):
    def __init__(self, validation_extensions_field_absent):
        self._validation_extensions_field_absent = validation_extensions_field_absent

        super().__init__(
            validations=validation_extensions_field_absent,
            pdu_class=rfc5280.TBSCertificate,
        )

    def validate(self, node):
        if "extensions" not in node.children:
            raise validation.ValidationFindingEncountered(
                self._validation_extensions_field_absent
            )


class ExtensionIdentifierAllowanceValidator(ElementIdentifierAllowanceValidator):
    @staticmethod
    def _retrieve_extension_id(node):
        return node.children["extnID"]

    def __init__(
        self,
        extension_allowances,
        finding_code_classifier: str,
        unknown_extension_allowance: Rfc2119Word,
    ):
        unknown_extension_finding = (
            None
            if unknown_extension_allowance == Rfc2119Word.MAY
            else (
                validation.ValidationFinding(
                    unknown_extension_allowance.to_severity,
                    f"{finding_code_classifier}.unknown_extension_present",
                )
            )
        )

        super().__init__(
            "extension",
            ExtensionIdentifierAllowanceValidator._retrieve_extension_id,
            extension_allowances,
            f"{finding_code_classifier}.{{oid}}_extension_present",
            f"{finding_code_classifier}.{{oid}}_extension_absent",
            unknown_extension_finding,
            pdu_class=rfc5280.Extensions,
        )


class ExtendedKeyUsageAllowanceValidator(ElementIdentifierAllowanceValidator):
    @staticmethod
    def _retrieve_eku(node):
        return node

    def __init__(
        self,
        eku_allowances,
        finding_code_classifier: str,
        unknown_eku_allowance: Rfc2119Word,
    ):
        unknown_eku_finding = (
            None
            if unknown_eku_allowance == Rfc2119Word.MAY
            else (
                validation.ValidationFinding(
                    unknown_eku_allowance.to_severity,
                    f"{finding_code_classifier}.unknown_eku_present",
                )
            )
        )
        super().__init__(
            "EKU",
            ExtendedKeyUsageAllowanceValidator._retrieve_eku,
            eku_allowances,
            f"{finding_code_classifier}.{{oid}}_eku_present",
            f"{finding_code_classifier}.{{oid}}_eku_absent",
            unknown_eku_finding,
            pdu_class=rfc5280.ExtKeyUsageSyntax,
        )


class ExtensionCriticalityValidator(validation.Validator):
    def __init__(
        self,
        criticality_mapping: Mapping[univ.ObjectIdentifier, bool],
        finding_code_classifier: str,
        critical_adherence_word: Rfc2119Word,
        non_critical_adherence_word: Rfc2119Word,
    ):
        self._expected_critical_extensions = {
            o: validation.ValidationFinding(
                critical_adherence_word.to_severity,
                f"{finding_code_classifier}.non_critical_{OID_TO_CODE_NAME[o]}_extension",
            )
            for o, c in criticality_mapping.items()
            if c and critical_adherence_word is not Rfc2119Word.MAY
        }
        self._expected_non_critical_extensions = {
            o: validation.ValidationFinding(
                critical_adherence_word.to_severity,
                f"{finding_code_classifier}.critical_{OID_TO_CODE_NAME[o]}_extension",
            )
            for o, c in criticality_mapping.items()
            if not c and non_critical_adherence_word is not Rfc2119Word.MAY
        }

        validations = list(self._expected_critical_extensions.values()) + list(
            self._expected_non_critical_extensions.values()
        )

        super().__init__(validations=validations, pdu_class=rfc5280.Extension)

    def validate(self, node):
        ext_oid = node.children["extnID"].pdu

        actual_criticality = bool(node.children["critical"].pdu)

        if actual_criticality:
            non_critical_finding = self._expected_non_critical_extensions.get(ext_oid)

            if non_critical_finding is not None:
                raise validation.ValidationFindingEncountered(non_critical_finding)
        else:
            critical_finding = self._expected_critical_extensions.get(ext_oid)

            if critical_finding is not None:
                raise validation.ValidationFindingEncountered(critical_finding)


class AuthorityInformationAccessAccessMethodPresenceValidator(
    ElementIdentifierAllowanceValidator
):
    @staticmethod
    def _retrieve_access_method_id(node):
        return node.children["accessMethod"]

    def __init__(
        self,
        access_method_allowances,
        finding_code_classifier: str,
        unknown_access_method_allowance: Rfc2119Word,
    ):
        unknown_access_method_finding = (
            None
            if unknown_access_method_allowance == Rfc2119Word.MAY
            else (
                validation.ValidationFinding(
                    unknown_access_method_allowance.to_severity,
                    f"{finding_code_classifier}.unknown_aia_access_method_present",
                )
            )
        )

        super().__init__(
            "access method",
            AuthorityInformationAccessAccessMethodPresenceValidator._retrieve_access_method_id,
            access_method_allowances,
            f"{finding_code_classifier}.{{oid}}_aia_access_method_present",
            f"{finding_code_classifier}.{{oid}}_aia_access_method_absent",
            unknown_access_method_finding,
            pdu_class=rfc5280.AuthorityInfoAccessSyntax,
        )


class AttributeIdentifierAllowanceValidator(ElementIdentifierAllowanceValidator):
    @staticmethod
    def _retrieve_attribute_type_id(node):
        # assume one ATV per RDN
        return node.navigate("0.type")

    def __init__(
        self,
        attribute_allowances,
        finding_code_classifier: str,
        unknown_attribute_allowance: Rfc2119Word,
    ):
        unexpected_attribute_finding = (
            None
            if unknown_attribute_allowance == Rfc2119Word.MAY
            else (
                validation.ValidationFinding(
                    unknown_attribute_allowance.to_severity,
                    finding_code_classifier + ".unknown_attribute_present",
                )
            )
        )

        super().__init__(
            "attribute",
            AttributeIdentifierAllowanceValidator._retrieve_attribute_type_id,
            attribute_allowances,
            finding_code_classifier + ".{oid}_attribute_present",
            finding_code_classifier + ".{oid}_attribute_absent",
            unexpected_attribute_finding,
            path="certificate.tbsCertificate.subject.rdnSequence",
        )



from pyasn1_alt_modules import rfc6960

from pkilint import document, validation


class OCSPResponseDecoder(document.ValueDecoder):
    def __init__(self, *, type_mappings):
        super().__init__(
            type_path="responseType", value_path="response", type_mappings=type_mappings
        )


class OCSPResponseDecodingValidator(validation.DecodingValidator):
    def __init__(self, *, decode_func, **kwargs):
        super().__init__(
            pdu_class=rfc6960.ResponseBytes, decode_func=decode_func, **kwargs
        )


class OCSPResponseStatusValidator(validation.Validator):
    VALIDATION_ERROR_RESPONSE_HAS_BYTES = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.ocsp_error_response_has_response_body",
    )

    VALIDATION_STATUS_NUMBER_4 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.ocsp_response_status_code_4"
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_ERROR_RESPONSE_HAS_BYTES,
                self.VALIDATION_STATUS_NUMBER_4,
            ],
            pdu_class=rfc6960.OCSPResponseStatus,
        )

    def validate(self, node):
        findings = []

        if node.pdu == rfc6960.OCSPResponseStatus("undefinedStatus"):
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_STATUS_NUMBER_4, None
                )
            )

        if (
            node.pdu != rfc6960.OCSPResponseStatus("successful")
            and "responseBytes" in node.parent.children
        ):
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_ERROR_RESPONSE_HAS_BYTES, None
                )
            )

        return validation.ValidationResult(self, node, findings)


class OCSPResponseIsBasicValidator(validation.Validator):
    VALIDATION_RESPONSE_IS_NOT_BASIC = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.ocsp_response_is_not_basic"
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_RESPONSE_IS_NOT_BASIC],
            pdu_class=rfc6960.ResponseBytes,
        )

    def validate(self, node):
        type_oid = node.children["responseType"].pdu

        if type_oid != rfc6960.id_pkix_ocsp_basic:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_RESPONSE_IS_NOT_BASIC,
                f"Unknown response type: {str(type_oid)}",
            )



from pyasn1.type import univ


_ARC = univ.ObjectIdentifier("0.4.0.2042.1")

id_ncp = _ARC + (1,)

id_ncp_plus = _ARC + (2,)

id_lcp = _ARC + (3,)

id_evcp = _ARC + (4,)

id_dvcp = _ARC + (6,)

id_ovcp = _ARC + (7,)

id_ivcp = _ARC + (8,)

POLICY_OIDS = {
    id_ncp,
    id_ncp_plus,
    id_lcp,
    id_evcp,
    id_dvcp,
    id_ovcp,
    id_ivcp,
}



from pkilint import validation, etsi
from pkilint.etsi import etsi_constants
from pkilint.pkix import certificate
from pkilint.cabf import serverauth, smime
import re


_FINDING_CODE_REGEX = re.compile(r"^[.a-z0-9\-_]+$")


def _test(validator, context: str):
    declared_validations = {
        getattr(validator, a) for a in dir(validator) if a.startswith("VALIDATION_")
    }

    reported_validations = set(validator.validations)

    missing_validations = declared_validations - reported_validations

    validator_name = f"{context}: {validator.__class__.__name__}"

    assert not any(missing_validations), (
        f"{validator_name} does not declare that it reports the following finding(s): "
        f"{missing_validations}"
    )

    incorrect_code_syntax_findings = [
        v.code for v in validator.validations if not _FINDING_CODE_REGEX.match(v.code)
    ]

    assert not any(
        incorrect_code_syntax_findings
    ), f"{validator_name} has validations with incorrect code syntax: {incorrect_code_syntax_findings}"

    if isinstance(validator, validation.ValidatorContainer):
        for v in validator.validators:
            _test(v, context)


def test_serverauth():
    for cert_type in serverauth.serverauth_constants.CertificateType:
        validator = certificate.create_pkix_certificate_validator_container(
            serverauth.create_decoding_validators(),
            serverauth.create_validators(cert_type),
        )

        context = f"cabf.serverauth.{cert_type}"

        _test(validator, context)


def test_smime():
    for validation_level in smime.smime_constants.ValidationLevel:
        for generation in smime.smime_constants.Generation:
            validator = certificate.create_pkix_certificate_validator_container(
                smime.create_decoding_validators(),
                smime.create_subscriber_validators(validation_level, generation),
            )

            context = f"cabf.smime.{validation_level}-{generation}"

            _test(validator, context)


def test_etsi():
    for certificate_type in etsi_constants.CertificateType:
        validator = certificate.create_pkix_certificate_validator_container(
            serverauth.create_decoding_validators(),
            etsi.create_validators(certificate_type),
        )

        context = f"etsi.{certificate_type}"

        _test(validator, context)



import operator
import re
from datetime import datetime, timedelta
from datetime import timezone

from dateutil.relativedelta import relativedelta
from pyasn1.type import useful
from pyasn1_alt_modules import rfc5280

from pkilint import validation
from pkilint.document import PDUNavigationFailedError

_REGEX_UTC_TIME = re.compile(
    r"^(?P<year>\d{2})(?P<month>\d{2})(?P<day>\d{2})"
    r"(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})Z$"
)

_REGEX_GENERALIZED_TIME = re.compile(
    r"^(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})"
    r"(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})Z$"
)

VALIDATION_INVALID_TIME_SYNTAX = validation.ValidationFinding(
    validation.ValidationFindingSeverity.ERROR, "pkix.invalid_time_syntax"
)


def _match_to_datetime(m):
    # check if UTCTime
    has_2_digit_year = len(m["year"]) == 2

    date_components = {k: int(v) for k, v in m.groupdict().items()}

    if has_2_digit_year:
        if date_components["year"] >= 50:
            date_components["year"] += 1900
        else:
            date_components["year"] += 2000

    return datetime(
        date_components["year"],
        date_components["month"],
        date_components["day"],
        date_components["hour"],
        date_components["minute"],
        date_components["second"],
        0,  # msec
        timezone.utc,
    )


def parse_generalizedtime(value):
    strval = str(value)

    m = _REGEX_GENERALIZED_TIME.match(strval)
    if m is None:
        raise ValueError(
            f'"{strval}" does not match GeneralizedTime regular '
            f'expression "{_REGEX_GENERALIZED_TIME.pattern}"'
        )

    return _match_to_datetime(m)


def parse_utctime(value):
    strval = str(value)

    m = _REGEX_UTC_TIME.match(strval)
    if m is None:
        raise ValueError(
            f'"{strval}" does not match UTCTime regular '
            f'expression "{_REGEX_UTC_TIME.pattern}"'
        )

    return _match_to_datetime(m)


def parse_time_node(value):
    if "generalTime" in value.children:
        return parse_generalizedtime(value.children["generalTime"].pdu)
    else:
        return parse_utctime(value.children["utcTime"].pdu)


class TimeCorrectEncodingValidator(validation.Validator):
    VALIDATION_WRONG_TIME_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.wrong_time_useful_type"
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.Time,
            validations=[
                self.VALIDATION_WRONG_TIME_TYPE,
            ],
        )

    def validate(self, node):
        try:
            parsed_datetime = parse_time_node(node)
        except ValueError:
            # the time syntax validator will report any errors
            return

        if (
            ("generalTime" in node.children)
            and (parsed_datetime.year >= 1950)
            and (parsed_datetime.year < 2050)
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_WRONG_TIME_TYPE,
                "Time values that contain a year value of "
                f'"{parsed_datetime.year}" must be encoded using UTCTime',
            )


def _parse_date(node):
    try:
        if isinstance(node.pdu, rfc5280.Time):
            return parse_time_node(node)
        elif isinstance(node.pdu, useful.GeneralizedTime):
            return parse_generalizedtime(str(node.pdu))
        elif isinstance(node.pdu, useful.UTCTime):
            return parse_utctime(str(node.pdu))
        else:
            raise ValueError("Unsupported ASN.1 time type")
    except ValueError as e:
        raise validation.ValidationFindingEncountered(
            VALIDATION_INVALID_TIME_SYNTAX, f"{node.name}: {str(e)}"
        )


class ValidityPeriodDifferenceValidator(validation.Validator):
    VALIDATION_PERIOD_END_VALIDITY_NODE_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.validity_period_end_value_missing",
    )

    def __init__(self, *, end_validity_node_retriever, validations, **kwargs):
        self._end_validity_node_retriever = end_validity_node_retriever

        super().__init__(
            validations=[
                VALIDATION_INVALID_TIME_SYNTAX,
                self.VALIDATION_PERIOD_END_VALIDITY_NODE_MISSING,
            ]
            + validations,
            **kwargs,
        )

    def validate(self, node):
        try:
            end_node = self._end_validity_node_retriever(node)
        except PDUNavigationFailedError:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PERIOD_END_VALIDITY_NODE_MISSING
            )

        start_datetime = _parse_date(node)
        end_datetime = _parse_date(end_node)

        self.validate_date_range(start_datetime, end_datetime)

    def validate_date_range(self, start_datetime, end_datetime):
        pass


class SaneValidityPeriodValidator(ValidityPeriodDifferenceValidator):
    def __init__(self, *, end_validity_node_retriever, validation, **kwargs):
        self._invalid_validity_period_validation = validation

        super().__init__(
            end_validity_node_retriever=end_validity_node_retriever,
            validations=[validation],
            **kwargs,
        )

    def validate_date_range(self, start_datetime, end_datetime):
        if start_datetime > end_datetime:
            raise validation.ValidationFindingEncountered(
                self._invalid_validity_period_validation,
                f'Start of validity period "{start_datetime}" is greater than '
                f'end of validity period "{end_datetime}"',
            )


class ValidityPeriodThresholdsValidator(ValidityPeriodDifferenceValidator):
    def __init__(
        self,
        *,
        end_validity_node_retriever,
        inclusive_second=False,
        validity_period_thresholds,
        **kwargs,
    ):
        validity_period_thresholds.sort(key=lambda v: v[2].severity)

        validations = [v[2] for v in validity_period_thresholds]

        self._validity_period_thresholds = validity_period_thresholds
        self._inclusive_second = inclusive_second

        super().__init__(
            end_validity_node_retriever=end_validity_node_retriever,
            validations=validations,
            **kwargs,
        )

    def validate_date_range(self, start_datetime, end_datetime):
        if self._inclusive_second:
            end_datetime += timedelta(seconds=1)
        validity = end_datetime - start_datetime

        for op, threshold_value, finding in self._validity_period_thresholds:
            if isinstance(threshold_value, relativedelta):
                is_valid = op(end_datetime, start_datetime + threshold_value)
            else:
                is_valid = op(validity, threshold_value)

            if not is_valid:
                if op in [operator.ge, operator.gt]:
                    raise validation.ValidationFindingEncountered(
                        finding,
                        f"Validity period of {validity} is below minimum "
                        f"value of {threshold_value}",
                    )
                else:
                    raise validation.ValidationFindingEncountered(
                        finding,
                        f"Validity period of {validity} exceeds maximum "
                        f"value of {threshold_value}",
                    )


class UtcTimeCorrectSyntaxValidator(validation.Validator):
    VALIDATION_INCORRECT_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.utctime_incorrect_syntax"
    )

    def __init__(self):
        super().__init__(
            pdu_class=useful.UTCTime, validations=[self.VALIDATION_INCORRECT_SYNTAX]
        )

    def validate(self, node):
        try:
            parse_utctime(node.pdu)
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INCORRECT_SYNTAX, str(e)
            )


class GeneralizedTimeCorrectSyntaxValidator(validation.Validator):
    VALIDATION_INCORRECT_SYNTAX = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.generalizedtime_incorrect_syntax",
    )

    def __init__(self):
        super().__init__(
            pdu_class=useful.GeneralizedTime,
            validations=[self.VALIDATION_INCORRECT_SYNTAX],
        )

    def validate(self, node):
        try:
            parse_generalizedtime(node.pdu)
        except ValueError as e:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_INCORRECT_SYNTAX, str(e)
            )



from pkilint import validation, common
from pkilint.etsi import etsi_constants
from pkilint.etsi.asn1 import en_319_412_5
from iso3166 import countries_by_alpha2
from iso4217 import Currency
from urllib.parse import urlparse
from pyasn1_alt_modules import rfc3739
from pkilint.pkix import extension, Rfc2119Word
import iso639


class QcCCLegislationCountryCodeValidator(validation.Validator):
    """EN 319 412-5 4.2.4.: QCStatement stating the country or set of countries under the legislation of which the
    certificate is issued as a qualified certificate. Constrained by ISO 3166-1 alpha-2 codes only. This Validator
    will check to see if there is a country code at all or if it is a valid code."""

    VALIDATION_ISO_COUNTRY_CODE_LIST_EMPTY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.2.4.iso_country_code_list_empty",
    )
    VALIDATION_ISO_COUNTRY_CODE_INVALID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.2.4.iso_country_code_invalid",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_ISO_COUNTRY_CODE_LIST_EMPTY,
                self.VALIDATION_ISO_COUNTRY_CODE_INVALID,
            ],
            pdu_class=en_319_412_5.QcCClegislation,
        )

    def validate(self, node):
        if not node.children:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ISO_COUNTRY_CODE_LIST_EMPTY
            )
        for children in node.children.values():
            country = str(children.pdu)
            if country not in countries_by_alpha2:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_ISO_COUNTRY_CODE_INVALID,
                    f'Invalid country code found: "{country}"',
                )


class QcEuRetentionPeriodValidator(validation.Validator):
    """EN 319 412-5 4.3.3 QCStatement indicating the duration of the retention period
    material information. This QCStatement declares a retention period for material information
    relevant to the use of and reliance of on a certificate, expressed as a number of years after the expiry
    date of the certificate. So in short anything greater will be 0 will be valid.
    """

    VALIDATION_RETENTION_PERIOD_NOT_POSITIVE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.3.3.retention_period_years_not_positive",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_RETENTION_PERIOD_NOT_POSITIVE],
            pdu_class=en_319_412_5.QcEuRetentionPeriod,
        )

    def validate(self, node):
        # noinspection PyTypeChecker
        valid_yrs = int(node.pdu)
        if not valid_yrs > 0:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_RETENTION_PERIOD_NOT_POSITIVE
            )


class QcEuPDSLanguageValidator(validation.Validator):
    """Content of the QcEuPDS statement, in accordance with Clause 4.3.4 of EN 319-412-5.
    Valid ISO 639-1 language code"""

    VALIDATION_ISO_LANGUAGE_CODE_INVALID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.3.4.iso_language_code_invalid",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_ISO_LANGUAGE_CODE_INVALID],
            pdu_class=en_319_412_5.PdsLocation,
        )

    def validate(self, node):
        language_code = str(node.children["language"].pdu).lower()

        try:
            iso639.Language.from_part1(language_code)
        except iso639.LanguageNotFoundError:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ISO_LANGUAGE_CODE_INVALID,
                f'Invalid language code found: "{language_code}"',
            )


class QcEuPDSHttpsURLValidator(validation.Validator):
    """Content of the QcEuPDS statement, in accordance with Clause 4.3.4 of EN 319-412-5.
    Validator to check if the URL has the 'https' scheme."""

    VALIDATION_URL_SCHEME_NOT_HTTPS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.3.4.url_scheme_not_https",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_URL_SCHEME_NOT_HTTPS],
            pdu_class=en_319_412_5.PdsLocation,
        )

    def validate(self, node):
        url_string = str(node.children["url"].pdu)
        parsed_url = urlparse(url_string)

        if parsed_url.scheme.lower() != "https":
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_URL_SCHEME_NOT_HTTPS,
                f'Non-HTTPS URL scheme found: "{parsed_url.scheme}"',
            )


class QcTypeValidator(validation.Validator):
    """EN 319 412-5 4.2.3 Declares that a certificate is issued as one and only one of the purposes
    of electronic signature, electronic seal or web site authentication. According to Stephen
    a qwac should never have seal or sign but may have psd2."""

    VALIDATION_QC_TYPE_MISMATCH = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.2.3.qc_type_mismatch",
    )

    VALIDATION_MULTIPLE_QC_TYPE_VALUES_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.2.3.multiple_qc_type_values_present",
    )

    VALIDATION_QC_TYPE_LIST_EMPTY = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.2.3.qc_type_list_empty",
    )

    def __init__(self, certificate_type):
        super().__init__(
            validations=[
                self.VALIDATION_QC_TYPE_MISMATCH,
                self.VALIDATION_QC_TYPE_LIST_EMPTY,
                self.VALIDATION_MULTIPLE_QC_TYPE_VALUES_PRESENT,
            ],
            pdu_class=en_319_412_5.QcType,
        )

        self._certificate_type = certificate_type

        if certificate_type in etsi_constants.WEB_AUTHENTICATION_CERTIFICATE_TYPES:
            self._expected_qc_type = en_319_412_5.id_etsi_qct_web
        else:
            self._expected_qc_type = None

    def validate(self, node):
        if not node.children.values():
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_QC_TYPE_LIST_EMPTY
            )

        if len(node.children.values()) != 1:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_MULTIPLE_QC_TYPE_VALUES_PRESENT
            )

        if self._expected_qc_type:
            _, qctype_value = node.child

            if qctype_value.pdu != self._expected_qc_type:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_QC_TYPE_MISMATCH,
                    f'Certificate type is "{self._certificate_type.to_option_str}" but QcType qualified '
                    f'statement contains "{qctype_value.pdu}"',
                )


class QcEuLimitValueValidator(validation.Validator):
    """
    This QCStatement declares a limitation on the value of transaction for which a certificate
    can be used.
    MonetaryValue:: == SEQUENCE {
        currency Iso4217CurrencyCode,
        amount INTEGER,
        exponent INTEGR
    }
    -- value = amount * 10^exponent
    Iso4217CurrencyCode:: = CHOICE {
        alphabetic PrintableString (Size (3)), -- Recommended
        numeric INTEGER (1..999) }
        -- Alphabetic or numeric currency code as defined in ISO 4217
        -- It is recommended that the Alphabetic form is used
    }
    QCS-4.3.2-01: The currency codes shall be defined in ISO 4217.
    QCS-4.3.2-02: The alphabetic form of currency codes should be used.

    Things to validate - valid iso 4217 currency code
                       - warning if the numeric code is used
                       - Positive amount and exponent value
    """

    VALIDATION_CURRENCY_CODE_INVALID = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.3.2.currency_code_invalid",
    )

    DISCOURAGED_VALIDATION_NUMERIC_CURRENCY_CODE_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "etsi.en_319_412_5.gen-4.3.2.discouraged_numeric_currency_code_present",
    )

    VALIDATION_AMOUNT_NEGATIVE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.3.2.amount_negative",
    )

    VALIDATION_EXPONENT_NEGATIVE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.gen-4.3.2.exponent_negative",
    )

    def __init__(self):
        self._alpha_codes = set()
        self._numeric_codes = set()

        for currency in Currency:
            self._alpha_codes.add(currency.code)
            self._numeric_codes.add(currency.number)

        # Let's remove the testing and unknown currency codes from the set.
        alpha_bad_codes = {"XTS", "XXX"}
        numeric_bad_codes = {"999", "963"}

        self._alpha_codes -= alpha_bad_codes
        self._numeric_codes -= numeric_bad_codes

        super().__init__(
            validations=[
                self.VALIDATION_CURRENCY_CODE_INVALID,
                self.DISCOURAGED_VALIDATION_NUMERIC_CURRENCY_CODE_PRESENT,
                self.VALIDATION_AMOUNT_NEGATIVE,
                self.VALIDATION_EXPONENT_NEGATIVE,
            ],
            pdu_class=en_319_412_5.MonetaryValue,
        )

    def validate(self, node):
        findings = []
        currency_code_type, iso_code = node.children["currency"].child
        if currency_code_type != "alphabetic":
            findings.append(
                validation.ValidationFindingDescription(
                    self.DISCOURAGED_VALIDATION_NUMERIC_CURRENCY_CODE_PRESENT, None
                )
            )
            iso_code = int(iso_code.pdu)
        else:
            iso_code = str(iso_code.pdu)

        if iso_code not in self._alpha_codes and iso_code not in self._numeric_codes:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_CURRENCY_CODE_INVALID, None
                )
            )
        if node.children["amount"].pdu < 0:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_AMOUNT_NEGATIVE, None
                )
            )
        if node.children["exponent"].pdu < 0:
            findings.append(
                validation.ValidationFindingEncountered(
                    self.VALIDATION_EXPONENT_NEGATIVE, None
                )
            )

        return validation.ValidationResult(self, node, findings)


class QcStatementsExtensionCriticalityValidator(
    extension.ExtensionCriticalityValidator
):
    """EN 319 412-5 QCS-4.1-02 The qcStatements extension shall not be marked as critical"""

    VALIDATION_QCSTATEMENTS_EXTENSION_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_5.qcs-4.1-02.qcstatements_extension_is_critical",
    )

    def __init__(self):
        super().__init__(
            type_oid=rfc3739.id_pe_qcStatements,
            is_critical=False,
            validation=self.VALIDATION_QCSTATEMENTS_EXTENSION_CRITICAL,
        )


class QcStatementIdentifierAllowanceValidator(
    common.ElementIdentifierAllowanceValidator
):
    """
    EN 319 412-5:
    QCS-5-01: EU qualified certificates shall include QCStatements in accordance with table 2
    """

    _CODE_CLASSIFIER = "etsi.en_319_412_5.qcs-5.01"

    # qualified statements
    _OID_TO_CODE_NAME = {
        en_319_412_5.id_etsi_qcs_QcCompliance: "qc_compliance",
        en_319_412_5.id_etsi_qcs_QcType: "qc_type",
        en_319_412_5.id_etsi_qcs_QcCClegislation: "qc_cc_legislation",
        en_319_412_5.id_etsi_qcs_QcSSCD: "qc_sscd",
    }

    @classmethod
    def retrieve_qualified_statement_id(cls, node):
        return node.children["statementId"]

    def __init__(self, certificate_type: etsi_constants.CertificateType):
        allowances = {}

        if certificate_type in etsi_constants.EU_QWAC_TYPES:
            allowances[en_319_412_5.id_etsi_qcs_QcCompliance] = Rfc2119Word.MUST
            allowances[en_319_412_5.id_etsi_qcs_QcCClegislation] = Rfc2119Word.MUST_NOT
            allowances[en_319_412_5.id_etsi_qcs_QcType] = Rfc2119Word.MUST
        elif certificate_type in etsi_constants.NON_EU_QWAC_TYPES:
            allowances[en_319_412_5.id_etsi_qcs_QcCompliance] = Rfc2119Word.MUST
            allowances[en_319_412_5.id_etsi_qcs_QcCClegislation] = Rfc2119Word.MUST

        if certificate_type in etsi_constants.QWAC_TYPES:
            allowances[en_319_412_5.id_etsi_qcs_QcSSCD] = Rfc2119Word.MUST_NOT

        super().__init__(
            "qualified statement",
            self.retrieve_qualified_statement_id,
            allowances,
            f"{self._CODE_CLASSIFIER}.{{oid}}_qualified_statement_present",
            f"{self._CODE_CLASSIFIER}.{{oid}}_qualified_statement_absent",
            None,
            pdu_class=rfc3739.QCStatements,
        )



import csv
import io
import json
from typing import Iterable, Optional, Any, List

from pkilint import validation
from pkilint.validation import (
    ValidationFindingSeverity,
    ValidationResult,
    ValidationFindingDescription,
)


class ReportGeneratorBase:
    def __init__(
        self,
        results: Iterable[ValidationResult],
        severity_threshold: Optional[ValidationFindingSeverity],
        report_context: Optional[Any] = None,
    ):
        self.results = results
        self.severity_threshold = severity_threshold
        self.report_context = report_context

    def get_finding_descriptions_for_result(self, result):
        return [
            f
            for f in result.finding_descriptions
            if self.severity_threshold is None
            or f.finding.severity <= self.severity_threshold
        ]

    def is_relevant_result(self, result):
        return self.severity_threshold is None or any(
            self.get_finding_descriptions_for_result(result)
        )

    def handle_result(self, result) -> Optional[Any]:
        pass

    def handle_finding_description(
        self,
        result: ValidationResult,
        finding_description: ValidationFindingDescription,
        result_context: Optional[Any],
    ):
        pass

    def generate(self):
        for result in self.results:
            result_context = self.handle_result(result)

            for finding_description in self.get_finding_descriptions_for_result(result):
                self.handle_finding_description(
                    result, finding_description, result_context
                )


class ReportGeneratorPlaintext(ReportGeneratorBase):
    def __init__(self, results, severity_threshold):
        super().__init__(results, severity_threshold, io.StringIO())

    def handle_result(self, result):
        if self.is_relevant_result(result):
            self.report_context.write(f"{result.validator} @ {result.node.path}\n")

    def handle_finding_description(
        self,
        result: ValidationResult,
        finding_description: ValidationFindingDescription,
        result_context: Optional[Any],
    ):
        self.report_context.write(f"    {finding_description}\n")

    def generate(self):
        super().generate()

        return self.report_context.getvalue()


class ReportGeneratorCsv(ReportGeneratorBase):
    _CSV_FIELDNAMES = ["node_path", "validator", "severity", "code", "message"]

    def __init__(self, results, severity_threshold, output_headers=True):
        self._output_io = io.StringIO()

        output_csv = csv.DictWriter(self._output_io, fieldnames=self._CSV_FIELDNAMES)

        if output_headers:
            output_csv.writeheader()

        super().__init__(results, severity_threshold, output_csv)

    def handle_finding_description(self, result, finding_description, result_context):
        row = {
            "node_path": result.node.path,
            "validator": str(result.validator),
            "severity": finding_description.finding.severity.name,
            "code": finding_description.finding.code,
            "message": (
                ""
                if finding_description.message is None
                else finding_description.message
            ),
        }

        self.report_context.writerow(row)

    def generate(self):
        super().generate()

        return self._output_io.getvalue()


class ReportGeneratorJson(ReportGeneratorBase):
    def __init__(self, results, severity_threshold):
        super().__init__(results, severity_threshold, [])

    def handle_result(self, result) -> Optional[Any]:
        if self.is_relevant_result(result):
            result_dict = {
                "node_path": result.node.path,
                "validator": str(result.validator),
                "finding_descriptions": [],
            }

            self.report_context.append(result_dict)

            return result_dict["finding_descriptions"]

    def handle_finding_description(
        self,
        result: ValidationResult,
        finding_description: ValidationFindingDescription,
        result_context: Optional[Any],
    ):
        result_context.append(
            {
                "severity": finding_description.finding.severity.name,
                "code": finding_description.finding.code,
                "message": finding_description.message,
            }
        )

    def generate(self):
        super().generate()

        return json.dumps({"results": self.report_context})


def get_findings_count(
    results: Iterable[ValidationResult],
    severity_threshold: ValidationFindingSeverity = None,
):
    findings = 0
    for result in results:
        for finding in result.finding_descriptions:
            if (
                severity_threshold is None
                or finding.finding.severity <= severity_threshold
            ):
                findings += 1

    return findings


def report_wrapper(report_generator_cls, *args, **kwargs):
    report_generator = report_generator_cls(*args, **kwargs)

    return report_generator.generate()


REPORT_FORMATS = {
    "TEXT": ReportGeneratorPlaintext,
    "CSV": ReportGeneratorCsv,
    "JSON": ReportGeneratorJson,
}

_VALIDATION_LIST_CSV_FIELDNAMES = ["severity", "code"]


def get_included_validations(*args) -> List[validation.ValidationFinding]:
    all_validations = set()
    for validator in args:
        all_validations.update(validator.validations)

    return sorted(all_validations, key=lambda v: f"{int(v.severity)}-{v.code}")


def report_included_validations(*args) -> str:
    s = io.StringIO()

    c = csv.DictWriter(s, fieldnames=_VALIDATION_LIST_CSV_FIELDNAMES)
    c.writeheader()

    validations = get_included_validations(*args)

    for v in validations:
        c.writerow({"severity": str(v.severity), "code": v.code})

    return s.getvalue()



# Auto-generated by asn1ate v.0.6.0 from ev_guidelines.asn1
# (last modified on 2020-12-23 17:42:25.113525)

from pyasn1.type import univ, char, namedtype, tag, constraint
from pyasn1.type.univ import ObjectIdentifier, Choice
from pyasn1_alt_modules import rfc5280


def _OID(*components):
    output = []
    for x in tuple(components):
        if isinstance(x, univ.ObjectIdentifier):
            output.extend(list(x))
        else:
            output.append(int(x))

    return univ.ObjectIdentifier(output)


class CABFOrganizationIdentifier(univ.Sequence):
    pass


CABFOrganizationIdentifier.componentType = namedtype.NamedTypes(
    namedtype.NamedType(
        "registrationSchemeIdentifier",
        char.PrintableString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(3, 3)
        ),
    ),
    namedtype.NamedType(
        "registrationCountry",
        char.PrintableString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(2, 2)
        ),
    ),
    namedtype.OptionalNamedType(
        "registrationStateOrProvince",
        char.PrintableString()
        .subtype(subtypeSpec=constraint.ValueSizeConstraint(0, 128))
        .subtype(implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)),
    ),
    namedtype.NamedType("registrationReference", char.UTF8String()),
)

id_CABFOrganizationIdentifier = _OID(2, 23, 140, 3, 1)

id_evat_jurisdiction_localityName = ObjectIdentifier("1.3.6.1.4.1.311.60.2.1.1")


class EVGJurisdictionLocalityName(Choice):
    pass


EVGJurisdictionLocalityName.componentType = namedtype.NamedTypes(
    namedtype.NamedType(
        "teletexString",
        char.TeletexString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_locality_name)
        ),
    ),
    namedtype.NamedType(
        "printableString",
        char.PrintableString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_locality_name)
        ),
    ),
    namedtype.NamedType(
        "universalString",
        char.UniversalString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_locality_name)
        ),
    ),
    namedtype.NamedType(
        "utf8String",
        char.UTF8String().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_locality_name)
        ),
    ),
    namedtype.NamedType(
        "bmpString",
        char.BMPString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_locality_name)
        ),
    ),
)

id_evat_jurisdiction_stateOrProvinceName = ObjectIdentifier("1.3.6.1.4.1.311.60.2.1.2")


class EVGJurisdictionStateOrProvinceName(Choice):
    pass


EVGJurisdictionStateOrProvinceName.componentType = namedtype.NamedTypes(
    namedtype.NamedType(
        "teletexString",
        char.TeletexString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_state_name)
        ),
    ),
    namedtype.NamedType(
        "printableString",
        char.PrintableString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_state_name)
        ),
    ),
    namedtype.NamedType(
        "universalString",
        char.UniversalString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_state_name)
        ),
    ),
    namedtype.NamedType(
        "utf8String",
        char.UTF8String().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_state_name)
        ),
    ),
    namedtype.NamedType(
        "bmpString",
        char.BMPString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, rfc5280.ub_state_name)
        ),
    ),
)

id_evat_jurisdiction_countryName = ObjectIdentifier("1.3.6.1.4.1.311.60.2.1.3")


class EVGJurisdictionCountryName(rfc5280.X520countryName):
    pass


ATTRIBUTE_TYPE_MAPPINGS = {
    id_evat_jurisdiction_countryName: EVGJurisdictionCountryName(),
    id_evat_jurisdiction_stateOrProvinceName: EVGJurisdictionStateOrProvinceName(),
    id_evat_jurisdiction_localityName: EVGJurisdictionLocalityName(),
}

EXTENSION_MAPPINGS = {
    id_CABFOrganizationIdentifier: CABFOrganizationIdentifier(),
}



from pyasn1.type import univ
from pyasn1.type.constraint import ValueSizeConstraint

from pkilint import document
from pkilint.nist.asn1 import csor


ML_KEM_512_PublicKeySize = 800
ML_KEM_768_PublicKeySize = 1184
ML_KEM_1024_PublicKeySize = 1568


class MlKem512PublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        ML_KEM_512_PublicKeySize, ML_KEM_512_PublicKeySize
    )


class MlKem768PublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        ML_KEM_768_PublicKeySize, ML_KEM_768_PublicKeySize
    )


class MlKem1024PublicKey(univ.OctetString):
    subtypeSpec = ValueSizeConstraint(
        ML_KEM_1024_PublicKeySize, ML_KEM_1024_PublicKeySize
    )


ALGORITHM_OID_TO_KEY_MAPPINGS = {
    csor.id_alg_ml_kem_512: MlKem512PublicKey(),
    csor.id_alg_ml_kem_768: MlKem768PublicKey(),
    csor.id_alg_ml_kem_1024: MlKem1024PublicKey(),
}

ALGORITHM_OID_TO_PARAMETER_MAPPINGS = {
    k: document.ValueDecoder.VALUE_NODE_ABSENT
    for k in ALGORITHM_OID_TO_KEY_MAPPINGS.keys()
}



import binascii
from typing import NamedTuple, Set, Optional

from cryptography import exceptions
from cryptography.hazmat.primitives import hashes
from pyasn1.codec.der.encoder import encode
from pyasn1.error import PyAsn1Error
from pyasn1.type import univ
from pyasn1_alt_modules import rfc5280, rfc8410, rfc3279, rfc5480

from pkilint import validation, util, document
from pkilint.itu import bitstring
from pkilint.nist.asn1 import csor
from pkilint.pkix.certificate.certificate_extension import KeyUsageBitName
from pkilint.pkix.key import verify_signature


class SubjectPublicKeyDecoder(document.ValueDecoder):
    def __init__(self, *, type_mappings):
        super().__init__(
            type_path="algorithm.algorithm",
            value_path="subjectPublicKey",
            type_mappings=type_mappings,
        )

    def decode_value(self, node, type_node, value_node, pdu_type):
        if isinstance(pdu_type, univ.OctetString):
            # map the BIT STRING into an OCTET STRING
            try:
                pdu = pdu_type.clone(value=value_node.pdu.asOctets())
            except PyAsn1Error as e:
                # bubble up any constraint violations
                raise document.SubstrateDecodingFailedError(
                    value_node.document, pdu_type, value_node, str(e)
                )

            return document.create_and_append_node_from_pdu(
                value_node.document, pdu, value_node
            )
        else:
            return super().decode_value(node, type_node, value_node, pdu_type)


class SubjectPublicKeyDecodingValidator(validation.DecodingValidator):
    def __init__(self, *, decode_func, **kwargs):
        super().__init__(
            pdu_class=rfc5280.SubjectPublicKeyInfo, decode_func=decode_func, **kwargs
        )


class SubjectPublicKeyParametersDecoder(document.ValueDecoder):
    def __init__(self, *, type_mappings):
        super().__init__(
            type_path="algorithm.algorithm",
            value_path="algorithm.parameters",
            type_mappings=type_mappings,
        )


class SubjectPublicKeyParametersDecodingValidator(validation.DecodingValidator):
    def __init__(self, *, decode_func, **kwargs):
        super().__init__(
            pdu_class=rfc5280.SubjectPublicKeyInfo, decode_func=decode_func, **kwargs
        )


class SubjectKeyIdentifierValidator(validation.Validator):
    VALIDATION_UNKNOWN_METHOD = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "pkix.unknown_subject_key_identifier_calculation_method",
    )

    # TODO: consider renaming the finding code after weighing risk of user breakage
    VALIDATION_METHOD_1 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.INFO,
        "pkix.subject_key_identifier_method_1_identified",
    )

    # TODO: consider renaming the finding code after weighing risk of user breakage
    VALIDATION_METHOD_2 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.INFO,
        "pkix.subject_key_identifier_method_2_identified",
    )

    VALIDATION_RFC7093_METHOD_1 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.INFO,
        "pkix.subject_key_identifier_rfc7093_method_1_identified",
    )

    VALIDATION_RFC7093_METHOD_2 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.INFO,
        "pkix.subject_key_identifier_rfc7093_method_2_identified",
    )

    VALIDATION_RFC7093_METHOD_3 = validation.ValidationFinding(
        validation.ValidationFindingSeverity.INFO,
        "pkix.subject_key_identifier_rfc7093_method_3_identified",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_UNKNOWN_METHOD,
                self.VALIDATION_METHOD_1,
                self.VALIDATION_METHOD_2,
                self.VALIDATION_RFC7093_METHOD_1,
                self.VALIDATION_RFC7093_METHOD_2,
                self.VALIDATION_RFC7093_METHOD_3,
            ],
            pdu_class=rfc5280.SubjectKeyIdentifier,
        )

    @staticmethod
    def _calculate_rfc5280_method2_id(sha1_hash):
        last_8_octets = bytearray(sha1_hash[12:])
        last_8_octets[0] = 0x40 | (last_8_octets[0] & 0xF)

        return bytes(last_8_octets)

    _RFC7093_HASH_CLS_TO_FINDINGS = {
        hashes.SHA256: VALIDATION_RFC7093_METHOD_1,
        hashes.SHA384: VALIDATION_RFC7093_METHOD_2,
        hashes.SHA512: VALIDATION_RFC7093_METHOD_3,
    }

    # TODO: support RFC 7093 method 4
    @staticmethod
    def _calculate_rfc7093_method_hash(public_key_octets, hash_cls):
        h = util.calculate_hash(public_key_octets, hash_cls())

        # leftmost 160 bits (i.e., 20 octets)
        return h[:20]

    def validate(self, node):
        public_key_node = node.document.root.navigate(
            "tbsCertificate.subjectPublicKeyInfo.subjectPublicKey"
        )

        public_key_octets = public_key_node.pdu.asOctets()

        identifier_octets = bytes(node.pdu)

        public_key_sha1 = util.calculate_sha1_hash(public_key_octets)

        if identifier_octets == public_key_sha1:
            finding = self.VALIDATION_METHOD_1
        elif (
            identifier_octets
            == SubjectKeyIdentifierValidator._calculate_rfc5280_method2_id(
                public_key_sha1
            )
        ):
            finding = self.VALIDATION_METHOD_2
        else:
            finding = next(
                (
                    f
                    for h, f in SubjectKeyIdentifierValidator._RFC7093_HASH_CLS_TO_FINDINGS.items()
                    if SubjectKeyIdentifierValidator._calculate_rfc7093_method_hash(
                        public_key_octets, h
                    )
                    == identifier_octets
                ),
                self.VALIDATION_UNKNOWN_METHOD,
            )

        raise validation.ValidationFindingEncountered(finding)


class SubjectSignatureVerificationValidator(validation.Validator):
    VALIDATION_SIGNATURE_MISMATCH = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.signature_verification_failed"
    )

    VALIDATION_UNSUPPORTED_ALGORITHM = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "pkix.unsupported_algorithm",
    )

    def __init__(self, *, tbs_node_retriever, **kwargs):
        super().__init__(
            validations=[
                self.VALIDATION_SIGNATURE_MISMATCH,
                self.VALIDATION_UNSUPPORTED_ALGORITHM,
            ],
            **kwargs,
        )

        self._tbs_node_retriever = tbs_node_retriever

    def validate(self, node):
        issuer_cert_doc = document.get_document_by_name(node, "issuer")

        public_key = issuer_cert_doc.public_key_object
        if public_key is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNSUPPORTED_ALGORITHM
            )

        subject_crypto_doc = node.document.cryptography_object

        tbs_octets = encode(self._tbs_node_retriever(node).pdu)

        try:
            if not verify_signature(
                public_key,
                tbs_octets,
                node.pdu.asOctets(),
                subject_crypto_doc.signature_hash_algorithm,
            ):
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_SIGNATURE_MISMATCH
                )
        except exceptions.UnsupportedAlgorithm:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNSUPPORTED_ALGORITHM
            )


class AllowedPublicKeyAlgorithmEncodingValidator(validation.Validator):
    def __init__(self, *, validation, allowed_encodings, **kwargs):
        super().__init__(validations=[validation], **kwargs)

        self._allowed_encodings = allowed_encodings

    def validate(self, node):
        encoded = encode(node.pdu)

        if encoded not in self._allowed_encodings:
            encoded_str = binascii.hexlify(encoded).decode("us-ascii")

            raise validation.ValidationFindingEncountered(
                self._validations[0], f"Prohibited encoding: {encoded_str}"
            )


class SpkiKeyUsageConsistencyValidator(validation.Validator):
    class KeyUsageBitsAndValidationFindingPair(NamedTuple):
        key_usage_bits: Set[str]
        validation_finding: validation.ValidationFinding

    class AlgorithmKeyUsageRequirement(NamedTuple):
        allowed: "SpkiKeyUsageConsistencyValidator.KeyUsageBitsAndValidationFindingPair"
        required: Optional[
            "SpkiKeyUsageConsistencyValidator.KeyUsageBitsAndValidationFindingPair"
        ]

    VALIDATION_RSA_PROHIBITED_KEY_USAGE_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.key_usage_value_prohibited_for_rsa",
    )

    VALIDATION_EC_PROHIBITED_KEY_USAGE_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.key_usage_value_prohibited_for_ec",
    )

    VALIDATION_EDWARDS_MISSING_REQUIRED_KEY_USAGE_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.key_usage_value_required_but_missing_for_edwards_curve",
    )

    VALIDATION_EDWARDS_PROHIBITED_KEY_USAGE_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.key_usage_value_prohibited_for_edwards_curve",
    )

    VALIDATION_MLDSA_PROHIBITED_KEY_USAGE_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.key_usage_value_prohibited_for_mldsa",
    )

    VALIDATION_HASH_MLDSA_PROHIBITED_KEY_USAGE_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.key_usage_value_prohibited_for_hash_mldsa",
    )

    VALIDATION_SLHDSA_PROHIBITED_KEY_USAGE_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.key_usage_value_prohibited_for_slhdsa",
    )

    VALIDATION_HASH_SLHDSA_PROHIBITED_KEY_USAGE_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.key_usage_value_prohibited_for_hash_slhdsa",
    )

    VALIDATION_MLKEM_PROHIBITED_KEY_USAGE_VALUE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.key_usage_value_prohibited_for_mlkem",
    )

    VALIDATION_UNSUPPORTED_PUBLIC_KEY_ALGORITHM = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "pkix.public_key_algorithm_unsupported",
    )

    _RSA = AlgorithmKeyUsageRequirement(
        allowed=KeyUsageBitsAndValidationFindingPair(
            # all bits are allowed except for keyAgreement, see RFC 4055 section 1.2
            key_usage_bits=KeyUsageBitName.all_bits() - {KeyUsageBitName.KEY_AGREEMENT},
            validation_finding=VALIDATION_RSA_PROHIBITED_KEY_USAGE_VALUE,
        ),
        required=None,
    )

    _EC = AlgorithmKeyUsageRequirement(
        allowed=KeyUsageBitsAndValidationFindingPair(
            # all bits are allowed except for keyEncipherment and dataEncipherment, see RFC 8813 section 3
            key_usage_bits=KeyUsageBitName.all_bits()
            - {KeyUsageBitName.KEY_ENCIPHERMENT, KeyUsageBitName.DATA_ENCIPHERMENT},
            validation_finding=VALIDATION_EC_PROHIBITED_KEY_USAGE_VALUE,
        ),
        required=None,
    )

    _EDWARDS_KEY_AGREEMENT = AlgorithmKeyUsageRequirement(
        allowed=KeyUsageBitsAndValidationFindingPair(
            # see RFC 9295, section 3
            key_usage_bits={
                KeyUsageBitName.KEY_AGREEMENT,
                KeyUsageBitName.DECIPHER_ONLY,
                KeyUsageBitName.ENCIPHER_ONLY,
            },
            validation_finding=VALIDATION_EDWARDS_PROHIBITED_KEY_USAGE_VALUE,
        ),
        required=KeyUsageBitsAndValidationFindingPair(
            # see RFC 9295, section 3
            key_usage_bits={
                KeyUsageBitName.KEY_AGREEMENT,
            },
            validation_finding=VALIDATION_EDWARDS_MISSING_REQUIRED_KEY_USAGE_VALUE,
        ),
    )

    _DIGITAL_SIGNATURE_ALGORITHM_BITS = {
        KeyUsageBitName.DIGITAL_SIGNATURE,
        KeyUsageBitName.NON_REPUDIATION,
        KeyUsageBitName.KEY_CERT_SIGN,
        KeyUsageBitName.CRL_SIGN,
    }

    _EDWARDS_DIGITAL_SIGNATURE = AlgorithmKeyUsageRequirement(
        allowed=KeyUsageBitsAndValidationFindingPair(
            # see RFC 9295, section 3
            key_usage_bits=_DIGITAL_SIGNATURE_ALGORITHM_BITS,
            validation_finding=VALIDATION_EDWARDS_PROHIBITED_KEY_USAGE_VALUE,
        ),
        required=None,
    )

    _MLDSA = AlgorithmKeyUsageRequirement(
        allowed=KeyUsageBitsAndValidationFindingPair(
            # see
            # https://www.ietf.org/archive/id/draft-ietf-lamps-dilithium-certificates-05.html#name-key-usage-bits
            key_usage_bits=_DIGITAL_SIGNATURE_ALGORITHM_BITS,
            validation_finding=VALIDATION_MLDSA_PROHIBITED_KEY_USAGE_VALUE,
        ),
        required=None,
    )

    _HASH_MLDSA = AlgorithmKeyUsageRequirement(
        allowed=KeyUsageBitsAndValidationFindingPair(
            # see
            # https://www.ietf.org/archive/id/draft-ietf-lamps-dilithium-certificates-05.html#name-key-usage-bits
            key_usage_bits={
                KeyUsageBitName.DIGITAL_SIGNATURE,
                KeyUsageBitName.NON_REPUDIATION,
            },
            validation_finding=VALIDATION_HASH_MLDSA_PROHIBITED_KEY_USAGE_VALUE,
        ),
        required=None,
    )

    _SLHDSA = AlgorithmKeyUsageRequirement(
        allowed=KeyUsageBitsAndValidationFindingPair(
            # see https://www.ietf.org/archive/id/draft-ietf-lamps-x509-slhdsa-03.html#name-key-usage-bits
            key_usage_bits=_DIGITAL_SIGNATURE_ALGORITHM_BITS,
            validation_finding=VALIDATION_SLHDSA_PROHIBITED_KEY_USAGE_VALUE,
        ),
        required=None,
    )

    _HASH_SLHDSA = AlgorithmKeyUsageRequirement(
        allowed=KeyUsageBitsAndValidationFindingPair(
            # see https://www.ietf.org/archive/id/draft-ietf-lamps-x509-slhdsa-03.html#name-key-usage-bits
            key_usage_bits={
                KeyUsageBitName.DIGITAL_SIGNATURE,
                KeyUsageBitName.NON_REPUDIATION,
            },
            validation_finding=VALIDATION_HASH_SLHDSA_PROHIBITED_KEY_USAGE_VALUE,
        ),
        required=None,
    )

    _MLKEM = AlgorithmKeyUsageRequirement(
        allowed=KeyUsageBitsAndValidationFindingPair(
            # see https://datatracker.ietf.org/doc/html/draft-ietf-lamps-kyber-certificates-07#section-3
            key_usage_bits={KeyUsageBitName.KEY_ENCIPHERMENT},
            validation_finding=VALIDATION_MLKEM_PROHIBITED_KEY_USAGE_VALUE,
        ),
        required=None,
    )

    _KEY_USAGE_VALUE_ALLOWANCES = {}

    def __init__(self):
        validations = {self.VALIDATION_UNSUPPORTED_PUBLIC_KEY_ALGORITHM}

        for alg_allowance in self._KEY_USAGE_VALUE_ALLOWANCES.values():
            validations.add(alg_allowance.allowed.validation_finding)

            if alg_allowance.required:
                validations.add(alg_allowance.required.validation_finding)

        self._KEY_USAGE_VALUE_ALLOWANCES.update(
            {
                rfc3279.rsaEncryption: self._RSA,
                rfc5480.id_ecPublicKey: self._EC,
                **{
                    k: self._EDWARDS_KEY_AGREEMENT
                    for k in (
                        rfc8410.id_X448,
                        rfc8410.id_X25519,
                    )
                },
                **{
                    k: self._EDWARDS_DIGITAL_SIGNATURE
                    for k in (
                        rfc8410.id_Ed448,
                        rfc8410.id_Ed25519,
                    )
                },
                **{k: self._MLDSA for k in csor.MLDSA_OIDS},
                **{k: self._HASH_MLDSA for k in csor.HASH_MLDSA_OIDS},
                **{k: self._SLHDSA for k in csor.SLHDSA_OIDS},
                **{k: self._HASH_SLHDSA for k in csor.HASH_SLHDSA_OIDS},
                **{k: self._MLKEM for k in csor.MLKEM_OIDS},
            }
        )

        super().__init__(
            validations=list(validations),
            pdu_class=rfc5280.KeyUsage,
        )

    def validate(self, node):
        spki_alg_oid = node.navigate(
            ":certificate.tbsCertificate.subjectPublicKeyInfo.algorithm.algorithm"
        ).pdu

        alg_allowances = self._KEY_USAGE_VALUE_ALLOWANCES.get(spki_alg_oid)

        if alg_allowances is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNSUPPORTED_PUBLIC_KEY_ALGORITHM,
                f"Unsupported public key algorithm: {str(spki_alg_oid)}",
            )

        allowed_values_and_finding, required_values_and_finding = alg_allowances
        allowed_values, prohibited_finding = allowed_values_and_finding

        bit_set = bitstring.get_asserted_bit_set(node)

        prohibited_bits = bit_set - allowed_values

        if any(prohibited_bits):
            prohibited_ku_names = ", ".join(sorted(prohibited_bits))

            raise validation.ValidationFindingEncountered(
                prohibited_finding,
                f"Prohibited key usage value(s) present: {prohibited_ku_names}",
            )

        if required_values_and_finding is not None:
            required_values, missing_finding = required_values_and_finding

            missing_kus = required_values - bit_set

            if any(missing_kus):
                missing_ku_names = ", ".join(sorted(missing_kus))

                raise validation.ValidationFindingEncountered(
                    missing_finding,
                    f"Required key usage value(s) missing: {missing_ku_names}",
                )


class CaPrehashPublicKeyValidator(validation.Validator):
    VALIDATION_HASH_MLDSA_PROHIBITED_IN_CA_CERTIFICATE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.hash_mldsa_ca_key_prohibited"
    )

    VALIDATION_HASH_SLHDSA_PROHIBITED_IN_CA_CERTIFICATE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.hash_slhdsa_ca_key_prohibited"
    )

    _PROHIBITED_ALG_OID_TO_FINDING_MAPPINGS = {}

    def __init__(self):
        self._PROHIBITED_ALG_OID_TO_FINDING_MAPPINGS.update(
            {
                **{
                    k: self.VALIDATION_HASH_MLDSA_PROHIBITED_IN_CA_CERTIFICATE
                    for k in csor.HASH_MLDSA_OIDS
                },
                **{
                    k: self.VALIDATION_HASH_SLHDSA_PROHIBITED_IN_CA_CERTIFICATE
                    for k in csor.HASH_SLHDSA_OIDS
                },
            }
        )

        super().__init__(
            validations=[
                self.VALIDATION_HASH_MLDSA_PROHIBITED_IN_CA_CERTIFICATE,
                self.VALIDATION_HASH_SLHDSA_PROHIBITED_IN_CA_CERTIFICATE,
            ],
            pdu_class=rfc5280.SubjectPublicKeyInfo,
            predicate=lambda n: n.document.is_ca,
        )

    def validate(self, node):
        spki_alg_oid = node.navigate("algorithm.algorithm").pdu

        finding = self._PROHIBITED_ALG_OID_TO_FINDING_MAPPINGS.get(spki_alg_oid)

        if finding:
            raise validation.ValidationFindingEncountered(
                finding,
                f"Prohibited public key algorithm in CA certificate: {str(spki_alg_oid)}",
            )


class ObsoletePublicKeyAlgorithmValidator(validation.Validator):
    VALIDATION_IPD_ALGORITHM_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.public_key_nist_ipd_algorithm_present",
    )

    VALIDATION_ROUND3_ALGORITHM_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "pkix.public_key_nist_round3_algorithm_present",
    )

    _IPD_ALG_OIDS = {
        univ.ObjectIdentifier("1.3.6.1.4.1.2.267.12.4.4"),  # ML-DSA-44-ipd
        univ.ObjectIdentifier("1.1.4.1.2.267.12.6.5"),  # ML-DSA-65-ipd
        univ.ObjectIdentifier("1.1.4.1.2.267.12.8.7"),  # ML-DSA-87-ipd
        univ.ObjectIdentifier("1.3.9999.6.4.16"),  # SLH-DSA-SHA2-128s-ipd
        univ.ObjectIdentifier("1.3.9999.6.7.16"),  # SLH-DSA-SHAKE-128s-ipd
        univ.ObjectIdentifier("1.3.9999.6.4.13"),  # SLH-DSA-SHA2-128f-ipd
        univ.ObjectIdentifier("1.3.9999.6.7.13"),  # SLH-DSA-SHAKE-128f-ipd
        univ.ObjectIdentifier("1.3.9999.6.5.12"),  # SLH-DSA-SHA2-192s-ipd
        univ.ObjectIdentifier("1.3.9999.6.8.12"),  # SLH-DSA-SHAKE-192s-ipd
        univ.ObjectIdentifier("1.3.9999.6.5.10"),  # SLH-DSA-SHA2-192f-ipd
        univ.ObjectIdentifier("1.3.9999.6.8.10"),  # SLH-DSA-SHAKE-192f-ipd
        univ.ObjectIdentifier("1.3.9999.6.6.12"),  # SLH-DSA-SHA2-256s-ipd
        univ.ObjectIdentifier("1.3.9999.6.9.12"),  # SLH-DSA-SHAKE-256s-ipd
        univ.ObjectIdentifier("1.3.9999.6.6.10"),  # SLH-DSA-SHA2-256f-ipd
        univ.ObjectIdentifier("1.3.9999.6.9.10"),  # SLH-DSA-SHAKE-256f-ipd
        univ.ObjectIdentifier("1.3.6.1.4.1.22554.5.6.1"),  # ML-KEM-512-ipd
        univ.ObjectIdentifier("1.3.6.1.4.1.22554.5.6.2"),  # ML-KEM-768-ipd
        univ.ObjectIdentifier("1.3.6.1.4.1.22554.5.6.3"),  # ML-KEM-1024-ipd
    }

    _ROUND3_ALG_OIDS = {
        univ.ObjectIdentifier("1.3.6.1.4.1.2.267.7.4.4"),  # Dilithium2
        univ.ObjectIdentifier("1.3.6.1.4.1.2.267.7.6.5"),  # Dilithium3
        univ.ObjectIdentifier("1.3.6.1.4.1.2.267.7.8.7"),  # Dilithium5
        univ.ObjectIdentifier("1.3.6.1.4.1.2.267.11.4.4"),  # DilithiumAES2
        univ.ObjectIdentifier("1.3.6.1.4.1.2.267.11.6.5"),  # DilithiumAES3
        univ.ObjectIdentifier("1.3.6.1.4.1.2.267.11.8.7"),  # DilithiumAES5
        univ.ObjectIdentifier("1.3.9999.3.1"),  # Falcon-512
        univ.ObjectIdentifier("1.3.9999.3.4"),  # Falcon-1024
        univ.ObjectIdentifier("1.3.9999.6.4.1"),  # SPHINCS+-SHA256-128f-robust
        univ.ObjectIdentifier("1.3.9999.6.4.4"),  # SPHINCS+-SHA256-128f-simple
        univ.ObjectIdentifier("1.3.9999.6.4.7"),  # SPHINCS+-SHA256-128s-robust
        univ.ObjectIdentifier("1.3.9999.6.4.10"),  # SPHINCS+-SHA256-128s-simple
        univ.ObjectIdentifier("1.3.9999.6.5.1"),  # SPHINCS+-SHA256-192f-robust
        univ.ObjectIdentifier("1.3.9999.6.5.3"),  # SPHINCS+-SHA256-192f-simple
        univ.ObjectIdentifier("1.3.9999.6.5.5"),  # SPHINCS+-SHA256-192s-robust
        univ.ObjectIdentifier("1.3.9999.6.5.7"),  # SPHINCS+-SHA256-192s-simple
        univ.ObjectIdentifier("1.3.9999.6.6.1"),  # SPHINCS+-SHA256-256f-robust
        univ.ObjectIdentifier("1.3.9999.6.6.3"),  # SPHINCS+-SHA256-256f-simple
        univ.ObjectIdentifier("1.3.9999.6.6.5"),  # SPHINCS+-SHA256-256s-robust
        univ.ObjectIdentifier("1.3.9999.6.6.7"),  # SPHINCS+-SHA256-256s-simple
        univ.ObjectIdentifier("1.3.6.1.4.1.22554.5.6.4"),  # kyber512_aes
        univ.ObjectIdentifier("1.3.6.1.4.1.22554.5.6.5"),  # kyber768_aes
        univ.ObjectIdentifier("1.3.6.1.4.1.22554.5.6.6"),  # kyber1024_aes
    }

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_IPD_ALGORITHM_PRESENT,
                self.VALIDATION_ROUND3_ALGORITHM_PRESENT,
            ],
            pdu_class=rfc5280.SubjectPublicKeyInfo,
        )

    def validate(self, node):
        spki_alg_oid = node.navigate("algorithm.algorithm").pdu

        if spki_alg_oid in self._IPD_ALG_OIDS:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_IPD_ALGORITHM_PRESENT,
                f"Obsolete NIST IPD public key algorithm: {str(spki_alg_oid)}",
            )

        if spki_alg_oid in self._ROUND3_ALG_OIDS:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_ROUND3_ALGORITHM_PRESENT,
                f"Obsolete NIST Round 3 public key algorithm: {str(spki_alg_oid)}",
            )



import datetime
from urllib.parse import urlparse

from pyasn1_alt_modules import rfc5280

from pkilint import validation, document


class CrlDpDistributionPointValidator(validation.Validator):
    """Validates that the fields included in the CRL distribution points extension conforms with BR 7.1.2.11.2."""

    VALIDATION_CRLDP_DP_PROHIBITED_FIELD = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.crldp_dp_prohibited_field_present",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_CRLDP_DP_PROHIBITED_FIELD,
            pdu_class=rfc5280.DistributionPoint,
        )

    def validate(self, node):
        for prohibited_field in ("reasons", "cRLIssuer"):
            if prohibited_field in node.children:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_CRLDP_DP_PROHIBITED_FIELD,
                    f"Prohibited field present: {prohibited_field}",
                )


# BR 7.1.2.11.2
class CrlDpDistributionPointNameValidator(validation.Validator):
    """Validates that the names included in the CRL distribution points extension conforms with BR 7.1.2.11.2."""

    VALIDATION_PROHIBITED_FIELD = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.crldp_dpname_prohibited_field_present",
    )

    VALIDATION_PROHIBITED_GENERALNAME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.crldp_dpname_prohibited_generalname_type",
    )

    VALIDATION_PROHIBITED_URI_SCHEME = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.crldp_dpname_prohibited_uri_scheme",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_PROHIBITED_FIELD,
                self.VALIDATION_PROHIBITED_GENERALNAME,
                self.VALIDATION_PROHIBITED_URI_SCHEME,
            ],
            pdu_class=rfc5280.DistributionPointName,
        )

    def validate(self, node):
        if "nameRelativeToCRLIssuer" in node.children:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_FIELD
            )

        gns = node.children["fullName"]

        prohibited_gn_types = {gn.child[0] for gn in gns.children.values()} - {
            "uniformResourceIdentifier"
        }

        if any(prohibited_gn_types):
            prohibited_types_str = ", ".join(prohibited_gn_types)

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_PROHIBITED_GENERALNAME,
                f"Prohibited GeneralName types present: {prohibited_types_str}",
            )

        for uri_node in (gn.child[1] for gn in gns.children.values()):
            scheme = urlparse(str(uri_node.pdu)).scheme

            if scheme.lower() != "http":
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_PROHIBITED_URI_SCHEME,
                    f'Prohibited URI scheme: "{scheme}"',
                )


class CrlDpDistributionPointCountValidator(validation.Validator):
    """Validates that the number of distribution points conforms with BR 7.1.2.11.2."""

    VALIDATION_MULTIPLE_CRLDP_DP = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.serverauth.crldp_multiple_distributionpoints_present",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_MULTIPLE_CRLDP_DP,
            pdu_class=rfc5280.CRLDistributionPoints,
        )

    def validate(self, node):
        if len(node.children) > 1:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_MULTIPLE_CRLDP_DP
            )


class AuthorityInformationAccessHttpUriLocationValidator(validation.Validator):
    """Validates that all locations in the AIA extension are HTTP as per BR 7.1.2.10.3, 7.1.2.7.7, and 7.1.2.8.3"""

    VALIDATION_AIA_LOCATION_NOT_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.aia_location_not_uri",
    )

    VALIDATION_AIA_LOCATION_URI_NOT_HTTP = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.aia_location_uri_not_http",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_AIA_LOCATION_NOT_URI,
                self.VALIDATION_AIA_LOCATION_URI_NOT_HTTP,
            ],
            pdu_class=rfc5280.AccessDescription,
        )

    def validate(self, node):
        gn_type, value = node.children["accessLocation"].child

        if gn_type != "uniformResourceIdentifier":
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AIA_LOCATION_NOT_URI,
                f"AIA access location is not URI: {gn_type}",
            )

        value_str = str(value.pdu)

        if not value_str.lower().startswith("http://"):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AIA_LOCATION_URI_NOT_HTTP,
                f'AIA access location is not HTTP: "{value_str}"',
            )


# BR 7.1.2.10.3, 7.1.2.7.7, 7.1.2.8.3
class AuthorityInformationAccessUniqueLocationValidator(validation.Validator):
    """Validates that all URI locations in the AIA extension are unique, as per BR 7.1.2.10.3, 7.1.2.7.7,
    and 7.1.2.8.3."""

    VALIDATION_DUPLICATE_LOCATION_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.aia_duplicate_location",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_DUPLICATE_LOCATION_URI,
            pdu_class=rfc5280.AuthorityInfoAccessSyntax,
        )

    def validate(self, node):
        uri_locations = []
        for access_description in node.children.values():
            location_type, location_value = access_description.children[
                "accessLocation"
            ].child

            if location_type == "uniformResourceIdentifier":
                uri_locations.append(str(location_value.pdu))

        duplicate_locations = {l for l in uri_locations if uri_locations.count(l) > 1}

        if any(duplicate_locations):
            dup_locations_str = ", ".join(sorted(list(duplicate_locations)))

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_DUPLICATE_LOCATION_URI,
                f'Duplicate AIA access locations: "{dup_locations_str}"',
            )


class CertificatePolicyQualifierValidator(validation.Validator):
    """Validates that the inclusion of policy qualifiers is in conformance with BR 7.1.2.3.2, 7.1.2.10.5,
    and 7.1.2.7.9."""

    VALIDATION_QUALIFIER_NOT_RECOMMENDED = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "cabf.serverauth.certificate_policy_qualifier_present",
    )

    VALIDATION_PROHIBITED_QUALIFIER_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.prohibited_certificate_policy_qualifier_type",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_QUALIFIER_NOT_RECOMMENDED,
                self.VALIDATION_PROHIBITED_QUALIFIER_TYPE,
            ],
            pdu_class=rfc5280.PolicyQualifierInfo,
        )

    def validate(self, node):
        findings = [
            validation.ValidationFindingDescription(
                self.VALIDATION_QUALIFIER_NOT_RECOMMENDED, None
            )
        ]

        qualifier_type_oid = node.children["policyQualifierId"].pdu

        if qualifier_type_oid != rfc5280.id_qt_cps:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_PROHIBITED_QUALIFIER_TYPE,
                    f"Prohibited qualifier type: {str(qualifier_type_oid)}",
                )
            )

        return validation.ValidationResult(self, node, findings)


class EvCpsUriPresenceValidator(validation.Validator):
    """Validates that EV Subscriber certificates contain the CPS URI qualifier, as per EVG 9.7."""

    VALIDATION_EV_CPS_URI_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.ev_guidelines.cps_uri_policy_qualifier_missing",
    )

    _VALIDATION_EV_CPS_URI_MISSING_INEFFECTIVE_DATE = datetime.datetime(
        2024, 5, 6, 0, 0, 0, tzinfo=datetime.timezone.utc
    )

    def __init__(
        self, validity_period_start_retriever: document.ValidityPeriodStartRetriever
    ):
        super().__init__(
            validations=self.VALIDATION_EV_CPS_URI_MISSING,
            pdu_class=rfc5280.CertificatePolicies,
        )

        self._validity_period_start_retriever = validity_period_start_retriever

    def match(self, node):
        return (
            super().match(node)
            and self._validity_period_start_retriever(node.document)
            < self._VALIDATION_EV_CPS_URI_MISSING_INEFFECTIVE_DATE
        )

    def validate(self, node):
        qualifier_oids = set()

        for pi in node.children.values():
            qualifiers = pi.children.get("policyQualifiers")

            if qualifiers is None:
                continue

            qualifier_oids.update(
                (
                    q.children["policyQualifierId"].pdu
                    for q in qualifiers.children.values()
                )
            )

        if rfc5280.id_qt_cps not in qualifier_oids:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_EV_CPS_URI_MISSING
            )
